import numpy as np
import pytest
from vispy import keys

from napari.components._interaction_box_constants import Box
from napari.utils.transforms import Affine


def check_corners_of_axis_aligned_interaction_box(
    box, *, top_left_corner=[0, 0], bottom_right_corner=None
):
    if not np.allclose(box[Box.TOP_LEFT], np.array(top_left_corner) - 0.5):
        pytest.fail(
            f"Top-left corner incorrect {box[Box.TOP_LEFT]} vs {np.array(top_left_corner) - 0.5}"
        )
    if not np.allclose(
        box[Box.BOTTOM_RIGHT], np.array(bottom_right_corner) - 0.5
    ):
        pytest.fail(
            f"Bottom-right corner incorrect {box[Box.BOTTOM_RIGHT]} vs {np.array(bottom_right_corner) - 0.5}"
        )


def test_interaction_box_display(make_napari_viewer):
    viewer = make_napari_viewer()

    data = np.random.random((2, 6, 30, 40))
    layer = viewer.add_image(data)

    layer.mode = 'transform'

    assert viewer.overlays.interaction_box.show
    check_corners_of_axis_aligned_interaction_box(
        viewer.overlays.interaction_box._box,
        top_left_corner=[0, 0],
        bottom_right_corner=[30, 40],
    )


def test_disable_with_3d(make_napari_viewer):
    viewer = make_napari_viewer()

    data = np.random.random((2, 6, 30, 40))
    layer = viewer.add_image(data)

    layer.mode = 'transform'
    viewer.dims.ndisplay = 3
    assert layer.mode == 'pan_zoom'
    with pytest.warns(UserWarning):
        layer.mode = 'transform'
    assert layer.mode == 'pan_zoom'


def test_disable_on_layer_cange(make_napari_viewer):
    viewer = make_napari_viewer()

    data = np.random.random((2, 6, 30, 40))
    layer = viewer.add_image(data)

    layer.mode = 'transform'
    viewer.add_image(data)
    assert viewer.overlays.interaction_box.show is False
    viewer.layers.selection.active = layer
    assert viewer.overlays.interaction_box.show is True


def test_interaction_box_dim_change(make_napari_viewer):

    viewer = make_napari_viewer()

    data = np.random.random((2, 6, 30, 40))
    layer = viewer.add_image(data)

    layer.mode = 'transform'

    viewer.dims._roll()

    check_corners_of_axis_aligned_interaction_box(
        viewer.overlays.interaction_box._box,
        top_left_corner=[0, 0],
        bottom_right_corner=[6, 30],
    )

    viewer.dims.transpose()

    check_corners_of_axis_aligned_interaction_box(
        viewer.overlays.interaction_box._box,
        top_left_corner=[0, 0],
        bottom_right_corner=[30, 6],
    )


def test_vertex_highlight(make_napari_viewer):
    viewer = make_napari_viewer()

    data = np.random.random((2, 6, 30, 40))
    layer = viewer.add_image(data)

    layer.mode = 'transform'

    viewer.overlays.interaction_box.selected_vertex = 9

    np.testing.assert_almost_equal(
        viewer.window._qt_viewer.interaction_box_visual.round_marker_node._data[
            'a_bg_color'
        ][
            0
        ][
            :-1
        ],
        viewer.window._qt_viewer.interaction_box_visual._highlight_color,
    )


def test_panzoom_on_space(make_napari_viewer):
    viewer = make_napari_viewer()
    view = viewer.window._qt_viewer
    data = np.random.random((2, 6, 30, 40))
    layer = viewer.add_image(data)

    layer.mode = 'transform'
    view.canvas.events.key_press(key=keys.Key('Space'))
    assert layer.mode == 'pan_zoom'
    assert viewer.overlays.interaction_box.show is False


def test_transform_coupling(make_napari_viewer):
    viewer = make_napari_viewer()

    data = np.random.random((2, 6, 30, 40))
    layer = viewer.add_image(data)
    layer.mode = 'transform'

    layer.affine = Affine(scale=[0.5, 0.5, 0.5, 0.5])
    np.testing.assert_almost_equal(
        viewer.overlays.interaction_box.transform.scale, [0.5, 0.5]
    )

    viewer.overlays.interaction_box.transform_drag = Affine(scale=[2.0, 2.0])
    np.testing.assert_almost_equal(layer.affine.scale, [0.5, 0.5, 2.0, 2.0])


def test_interaction_box_changes_with_layer_transform(make_napari_viewer):
    viewer = make_napari_viewer()
    layer = viewer.add_image(np.random.random((28, 28)))
    layer.mode = 'transform'
    initial_selection_box = np.copy(
        viewer.overlays.interaction_box.transform.scale
    )
    layer.scale = [5, 5]
    np.testing.assert_almost_equal(
        initial_selection_box * 5,
        viewer.overlays.interaction_box.transform.scale,
    )



from functools import partial


def _load_skimage_data(name, **kwargs):
    import skimage.data

    if name == 'cells3d':
        return [
            (
                skimage.data.cells3d(),
                {
                    'channel_axis': 1,
                    'name': ['membrane', 'nuclei'],
                    'contrast_limits': [(1110, 23855), (1600, 50000)],
                },
            )
        ]
    elif name == 'kidney':
        return [
            (
                skimage.data.kidney(),
                {
                    'channel_axis': -1,
                    'name': ['nuclei', 'WGA', 'actin'],
                    'colormap': ['blue', 'green', 'red'],
                },
            )
        ]
    elif name == 'lily':
        return [
            (
                skimage.data.lily(),
                {
                    'channel_axis': -1,
                    'name': ['lily-R', 'lily-G', 'lily-W', 'lily-B'],
                    'colormap': ['red', 'green', 'gray', 'blue'],
                },
            )
        ]
    elif name == 'binary_blobs_3D':
        kwargs['n_dim'] = 3
        kwargs.setdefault('length', 128)
        kwargs.setdefault('volume_fraction', 0.25)
        name = 'binary_blobs'

    return [(getattr(skimage.data, name)(**kwargs), {'name': name})]


# fmt: off
SKIMAGE_DATA = [
    'astronaut', 'binary_blobs', 'binary_blobs_3D', 'brain', 'brick', 'camera', 'cat',
    'cell', 'cells3d', 'checkerboard', 'clock', 'coffee', 'coins', 'colorwheel',
    'eagle', 'grass', 'gravel', 'horse', 'hubble_deep_field', 'human_mitosis',
    'immunohistochemistry', 'kidney', 'lfw_subset', 'lily', 'microaneurysms', 'moon',
    'page', 'retina', 'rocket', 'shepp_logan_phantom', 'skin', 'text',
]

globals().update({key: partial(_load_skimage_data, key) for key in SKIMAGE_DATA})



from __future__ import annotations

import weakref
from contextlib import contextmanager
from contextvars import ContextVar
from typing import Optional, Tuple

from magicgui.widgets import FunctionGui
from pydantic import BaseModel, validator

from .base.base import Layer


class Source(BaseModel):
    """An object to store the provenance of a layer.

    Parameters
    ----------
    path: str, optional
        filpath/url associated with layer
    reader_plugin: str, optional
        name of reader plugin that loaded the file (if applicable)
    sample: Tuple[str, str], optional
        Tuple of (sample_plugin, sample_name), if layer was loaded via
        `viewer.open_sample`.
    widget: FunctionGui, optional
        magicgui widget, if the layer was added via a magicgui widget.
    parent: Layer, optional
        parent layer if the layer is a duplicate.
    """

    path: Optional[str] = None
    reader_plugin: Optional[str] = None
    sample: Optional[Tuple[str, str]] = None
    widget: Optional[FunctionGui] = None
    parent: Optional[Layer] = None

    class Config:
        arbitrary_types_allowed = True
        frozen = True

    @validator('parent')
    def make_weakref(cls, layer: Layer):
        return weakref.ref(layer)

    def __deepcopy__(self, memo):
        """Custom deepcopy implementation.

        this prevents deep copy. `Source` doesn't really need to be copied
        (i.e. if we deepcopy a layer, it essentially has the same `Source`).
        Moreover, deepcopying a widget is challenging, and maybe odd anyway.
        """
        return self


# layer source context management

_LAYER_SOURCE: ContextVar[dict] = ContextVar('_LAYER_SOURCE', default={})


@contextmanager
def layer_source(**source_kwargs):
    """Creates context in which all layers will be given `source_kwargs`.

    The module-level variable `_LAYER_SOURCE` holds a set of key-value pairs
    that can be used to create a new `Source` object.  Any routine in napari
    that may result in the creation of a new layer (such as opening a file,
    using a particular plugin, or calling a magicgui widget) can use this
    context manager to declare that any layers created within the context
    result from a specific source. (This applies even if the layer
    isn't "directly" created in the context, but perhaps in some sub-function
    within the context).

    `Layer.__init__` will call :func:`current_source`, to query the current
    state of the `_LAYER_SOURCE` variable.

    Contexts may be stacked, meaning a given layer.source can reflect the
    actions of multiple events (for instance, an `open_sample` call that in
    turn resulted in a `reader_plugin` opening a file).  However, the "deepest"
    context will "win" in the case where multiple calls to `layer_source`
    provide conflicting values.

    Parameters
    ----------
    **source_kwargs
        keys/values should be valid parameters for :class:`Source`.

    Examples
    --------

    >>> with layer_source(path='file.ext', reader_plugin='plugin'):  # doctest: +SKIP
    ...     points = some_function_that_creates_points()
    ...
    >>> assert points.source == Source(path='file.ext', reader_plugin='plugin')  # doctest: +SKIP

    """
    token = _LAYER_SOURCE.set({**_LAYER_SOURCE.get(), **source_kwargs})
    try:
        yield
    finally:
        _LAYER_SOURCE.reset(token)


def current_source():
    """Get the current layer :class:`Source` (inferred from context).

    The main place this function is used is in :meth:`Layer.__init__`.
    """
    return Source(**_LAYER_SOURCE.get())



"""Methods to create a new viewer instance then add a particular layer type.

All functions follow this pattern, (where <layer_type> is replaced with one
of the layer types, like "image", "points", etc...):

.. code-block:: python

    def view_<layer_type>(*args, **kwargs):
        # ... pop all of the viewer kwargs out of kwargs into viewer_kwargs
        viewer = Viewer(**viewer_kwargs)
        add_method = getattr(viewer, f"add_{<layer_type>}")
        add_method(*args, **kwargs)
        return viewer
"""
import inspect
from typing import Any, List, Tuple

from numpydoc.docscrape import NumpyDocString as _NumpyDocString

from napari.components.dims import Dims
from napari.layers import Image

from .viewer import Viewer

__all__ = [
    'view_image',
    'view_labels',
    'view_path',
    'view_points',
    'view_shapes',
    'view_surface',
    'view_tracks',
    'view_vectors',
    'imshow',
]

_doc_template = """Create a viewer and add a{n} {layer_string} layer.

{params}

Returns
-------
viewer : :class:`napari.Viewer`
    The newly-created viewer.
"""

_VIEW_DOC = _NumpyDocString(Viewer.__doc__)
_VIEW_PARAMS = "    " + "\n".join(_VIEW_DOC._str_param_list('Parameters')[2:])


def _merge_docstrings(add_method, layer_string):
    # create combined docstring with parameters from add_* and Viewer methods
    import textwrap

    add_method_doc = _NumpyDocString(add_method.__doc__)

    # this ugliness is because the indentation of the parsed numpydocstring
    # is different for the first parameter :(
    lines = add_method_doc._str_param_list('Parameters')
    lines = lines[:3] + textwrap.dedent("\n".join(lines[3:])).splitlines()
    params = "\n".join(lines) + "\n" + textwrap.dedent(_VIEW_PARAMS)
    n = 'n' if layer_string.startswith(tuple('aeiou')) else ''
    return _doc_template.format(n=n, layer_string=layer_string, params=params)


def _merge_layer_viewer_sigs_docs(func):
    """Make combined signature, docstrings, and annotations for `func`.

    This is a decorator that combines information from `Viewer.__init__`,
    and one of the `viewer.add_*` methods.  It updates the docstring,
    signature, and type annotations of the decorated function with the merged
    versions.

    Parameters
    ----------
    func : callable
        `view_<layer_type>` function to modify

    Returns
    -------
    func : callable
        The same function, with merged metadata.
    """
    from .utils.misc import _combine_signatures

    # get the `Viewer.add_*` method
    layer_string = func.__name__.replace("view_", "")
    if layer_string == 'path':
        add_method = Viewer.open
    else:
        add_method = getattr(Viewer, f'add_{layer_string}')

    # merge the docstrings of Viewer and viewer.add_*
    func.__doc__ = _merge_docstrings(add_method, layer_string)

    # merge the signatures of Viewer and viewer.add_*
    func.__signature__ = _combine_signatures(
        add_method, Viewer, return_annotation=Viewer, exclude=('self',)
    )

    # merge the __annotations__
    func.__annotations__ = {
        **add_method.__annotations__,
        **Viewer.__init__.__annotations__,
        'return': Viewer,
    }

    # _forwardrefns_ is used by stubgen.py to populate the globalns
    # when evaluate forward references with get_type_hints
    func._forwardrefns_ = {**add_method.__globals__}
    return func


_viewer_params = inspect.signature(Viewer).parameters
_dims_params = Dims.__fields__


def _make_viewer_then(
    add_method: str,
    args,
    kwargs,
) -> Tuple[Viewer, Any]:
    """Create a viewer, call given add_* method, then return viewer and layer.

    This function will be deprecated soon (See #4693)

    Parameters
    ----------
    add_method : str
        Which ``add_`` method to call on the viewer, e.g. `add_image`,
        or `add_labels`.
    args : list
        Positional arguments for the ``add_`` method.
    kwargs : dict
        Keyword arguments for either the `Viewer` constructor or for the
        ``add_`` method.
    viewer : Viewer, optional
        A pre-existing viewer, which will be used provided, rather than
        creating a new one.

    Returns
    -------
    viewer : napari.Viewer
        The created viewer, or the same one that was passed in, if given.
    layer(s): napari.layers.Layer or List[napari.layers.Layer]
        The value returned by the add_method. Can be a list of layers if
        ``add_image`` is called with a ``channel_axis=`` keyword
        argument.
    """
    vkwargs = {k: kwargs.pop(k) for k in list(kwargs) if k in _viewer_params}
    # separate dims kwargs because we want to set those after adding data
    dims_kwargs = {
        k: vkwargs.pop(k) for k in list(vkwargs) if k in _dims_params
    }
    viewer = kwargs.pop("viewer", None)
    if viewer is None:
        viewer = Viewer(**vkwargs)
    kwargs.update(kwargs.pop("kwargs", {}))
    method = getattr(viewer, add_method)
    added = method(*args, **kwargs)
    if isinstance(added, list):
        added = tuple(added)
    for arg_name, arg_val in dims_kwargs.items():
        setattr(viewer.dims, arg_name, arg_val)
    return viewer, added


# Each of the following functions will have this pattern:
#
# def view_image(*args, **kwargs):
#     # ... pop all of the viewer kwargs out of kwargs into viewer_kwargs
#     viewer = Viewer(**viewer_kwargs)
#     viewer.add_image(*args, **kwargs)
#     return viewer


@_merge_layer_viewer_sigs_docs
def view_image(*args, **kwargs):
    return _make_viewer_then('add_image', args, kwargs)[0]


@_merge_layer_viewer_sigs_docs
def view_labels(*args, **kwargs):
    return _make_viewer_then('add_labels', args, kwargs)[0]


@_merge_layer_viewer_sigs_docs
def view_points(*args, **kwargs):
    return _make_viewer_then('add_points', args, kwargs)[0]


@_merge_layer_viewer_sigs_docs
def view_shapes(*args, **kwargs):
    return _make_viewer_then('add_shapes', args, kwargs)[0]


@_merge_layer_viewer_sigs_docs
def view_surface(*args, **kwargs):
    return _make_viewer_then('add_surface', args, kwargs)[0]


@_merge_layer_viewer_sigs_docs
def view_tracks(*args, **kwargs):
    return _make_viewer_then('add_tracks', args, kwargs)[0]


@_merge_layer_viewer_sigs_docs
def view_vectors(*args, **kwargs):
    return _make_viewer_then('add_vectors', args, kwargs)[0]


@_merge_layer_viewer_sigs_docs
def view_path(*args, **kwargs):
    return _make_viewer_then('open', args, kwargs)[0]


def imshow(
    data,
    *,
    channel_axis=None,
    rgb=None,
    colormap=None,
    contrast_limits=None,
    gamma=1,
    interpolation2d='nearest',
    interpolation3d='linear',
    rendering='mip',
    depiction='volume',
    iso_threshold=0.5,
    attenuation=0.05,
    name=None,
    metadata=None,
    scale=None,
    translate=None,
    rotate=None,
    shear=None,
    affine=None,
    opacity=1,
    blending=None,
    visible=True,
    multiscale=None,
    cache=True,
    plane=None,
    experimental_clipping_planes=None,
    viewer=None,
    title='napari',
    ndisplay=2,
    order=(),
    axis_labels=(),
    show=True,
) -> Tuple[Viewer, List[Image]]:
    """Load data into an Image layer and return the Viewer and Layer.

    Parameters
    ----------
    data : array or list of array
        Image data. Can be N >= 2 dimensional. If the last dimension has length
        3 or 4 can be interpreted as RGB or RGBA if rgb is `True`. If a
        list and arrays are decreasing in shape then the data is treated as
        a multiscale image. Please note multiscale rendering is only
        supported in 2D. In 3D, only the lowest resolution scale is
        displayed.
    channel_axis : int, optional
        Axis to expand image along.  If provided, each channel in the data
        will be added as an individual image layer.  In channel_axis mode,
        all other parameters MAY be provided as lists, and the Nth value
        will be applied to the Nth channel in the data.  If a single value
        is provided, it will be broadcast to all Layers.
    rgb : bool or list
        Whether the image is rgb RGB or RGBA. If not specified by user and
        the last dimension of the data has length 3 or 4 it will be set as
        `True`. If `False` the image is interpreted as a luminance image.
        If a list then must be same length as the axis that is being
        expanded as channels.
    colormap : str, napari.utils.Colormap, tuple, dict, list
        Colormaps to use for luminance images. If a string must be the name
        of a supported colormap from vispy or matplotlib. If a tuple the
        first value must be a string to assign as a name to a colormap and
        the second item must be a Colormap. If a dict the key must be a
        string to assign as a name to a colormap and the value must be a
        Colormap. If a list then must be same length as the axis that is
        being expanded as channels, and each colormap is applied to each
        new image layer.
    contrast_limits : list (2,)
        Color limits to be used for determining the colormap bounds for
        luminance images. If not passed is calculated as the min and max of
        the image. If list of lists then must be same length as the axis
        that is being expanded and then each colormap is applied to each
        image.
    gamma : list, float
        Gamma correction for determining colormap linearity. Defaults to 1.
        If a list then must be same length as the axis that is being
        expanded as channels.
    interpolation : str or list
        Deprecated, to be removed in 0.6.0
    interpolation2d : str or list
        Interpolation mode used by vispy in 2D. Must be one of our supported
        modes. If a list then must be same length as the axis that is being
        expanded as channels.
    interpolation3d : str or list
        Interpolation mode used by vispy in 3D. Must be one of our supported
        modes. If a list then must be same length as the axis that is being
        expanded as channels.
    rendering : str or list
        Rendering mode used by vispy. Must be one of our supported
        modes. If a list then must be same length as the axis that is being
        expanded as channels.
    depiction : str
        Selects a preset volume depiction mode in vispy

        * volume: images are rendered as 3D volumes.
        * plane: images are rendered as 2D planes embedded in 3D.
    iso_threshold : float or list
        Threshold for isosurface. If a list then must be same length as the
        axis that is being expanded as channels.
    attenuation : float or list
        Attenuation rate for attenuated maximum intensity projection. If a
        list then must be same length as the axis that is being expanded as
        channels.
    name : str or list of str
        Name of the layer.  If a list then must be same length as the axis
        that is being expanded as channels.
    metadata : dict or list of dict
        Layer metadata. If a list then must be a list of dicts with the
        same length as the axis that is being expanded as channels.
    scale : tuple of float or list
        Scale factors for the layer. If a list then must be a list of
        tuples of float with the same length as the axis that is being
        expanded as channels.
    translate : tuple of float or list
        Translation values for the layer. If a list then must be a list of
        tuples of float with the same length as the axis that is being
        expanded as channels.
    rotate : float, 3-tuple of float, n-D array or list.
        If a float convert into a 2D rotation matrix using that value as an
        angle. If 3-tuple convert into a 3D rotation matrix, using a yaw,
        pitch, roll convention. Otherwise assume an nD rotation. Angles are
        assumed to be in degrees. They can be converted from radians with
        np.degrees if needed. If a list then must have same length as
        the axis that is being expanded as channels.
    shear : 1-D array or list.
        A vector of shear values for an upper triangular n-D shear matrix.
        If a list then must have same length as the axis that is being
        expanded as channels.
    affine : n-D array or napari.utils.transforms.Affine
        (N+1, N+1) affine transformation matrix in homogeneous coordinates.
        The first (N, N) entries correspond to a linear transform and
        the final column is a length N translation vector and a 1 or a
        napari `Affine` transform object. Applied as an extra transform on
        top of the provided scale, rotate, and shear values.
    opacity : float or list
        Opacity of the layer visual, between 0.0 and 1.0.  If a list then
        must be same length as the axis that is being expanded as channels.
    blending : str or list
        One of a list of preset blending modes that determines how RGB and
        alpha values of the layer visual get mixed. Allowed values are
        {'opaque', 'translucent', and 'additive'}. If a list then
        must be same length as the axis that is being expanded as channels.
    visible : bool or list of bool
        Whether the layer visual is currently being displayed.
        If a list then must be same length as the axis that is
        being expanded as channels.
    multiscale : bool
        Whether the data is a multiscale image or not. Multiscale data is
        represented by a list of array like image data. If not specified by
        the user and if the data is a list of arrays that decrease in shape
        then it will be taken to be multiscale. The first image in the list
        should be the largest. Please note multiscale rendering is only
        supported in 2D. In 3D, only the lowest resolution scale is
        displayed.
    cache : bool
        Whether slices of out-of-core datasets should be cached upon
        retrieval. Currently, this only applies to dask arrays.
    plane : dict or SlicingPlane
        Properties defining plane rendering in 3D. Properties are defined in
        data coordinates. Valid dictionary keys are
        {'position', 'normal', 'thickness', and 'enabled'}.
    experimental_clipping_planes : list of dicts, list of ClippingPlane, or ClippingPlaneList
        Each dict defines a clipping plane in 3D in data coordinates.
        Valid dictionary keys are {'position', 'normal', and 'enabled'}.
        Values on the negative side of the normal are discarded if the plane is enabled.
    viewer: Viewer object, optional, by default None.
    title : string, optional
        The title of the viewer window. By default 'napari'.
    ndisplay : {2, 3}, optional
        Number of displayed dimensions. By default 2.
    order : tuple of int, optional
        Order in which dimensions are displayed where the last two or last
        three dimensions correspond to row x column or plane x row x column if
        ndisplay is 2 or 3. By default None
    axis_labels : list of str, optional
        Dimension names. By default they are labeled with sequential numbers
    show : bool, optional
        Whether to show the viewer after instantiation. By default True.

    Returns
    -------
    viewer : napari.Viewer
        The created or passed viewer.
    layer(s) : napari.layers.Image or List[napari.layers.Image]
        The added layer(s). (May be more than one if the ``channel_axis`` keyword
        argument is given.
    """

    kwargs = dict(
        viewer=viewer,
        channel_axis=channel_axis,
        rgb=rgb,
        colormap=colormap,
        contrast_limits=contrast_limits,
        gamma=gamma,
        interpolation2d=interpolation2d,
        interpolation3d=interpolation3d,
        rendering=rendering,
        depiction=depiction,
        iso_threshold=iso_threshold,
        attenuation=attenuation,
        name=name,
        metadata=metadata,
        scale=scale,
        translate=translate,
        rotate=rotate,
        shear=shear,
        affine=affine,
        opacity=opacity,
        blending=blending,
        visible=visible,
        multiscale=multiscale,
        cache=cache,
        plane=plane,
        experimental_clipping_planes=experimental_clipping_planes,
        title=title,
        ndisplay=ndisplay,
        order=order,
        axis_labels=axis_labels,
        show=show,
    )

    args = tuple([data])

    viewer, layers = _make_viewer_then(
        'add_image',
        args,
        kwargs,
    )

    return viewer, layers



import numpy as np
import pytest

from napari.utils.geometry import (
    bounding_box_to_face_vertices,
    clamp_point_to_bounding_box,
    distance_between_point_and_line_3d,
    face_coordinate_from_bounding_box,
    find_front_back_face,
    find_nearest_triangle_intersection,
    inside_triangles,
    intersect_line_with_axis_aligned_bounding_box_3d,
    intersect_line_with_axis_aligned_plane,
    intersect_line_with_multiple_planes_3d,
    intersect_line_with_plane_3d,
    line_in_quadrilateral_3d,
    line_in_triangles_3d,
    point_in_quadrilateral_2d,
    project_points_onto_plane,
    rotation_matrix_from_vectors_2d,
    rotation_matrix_from_vectors_3d,
)

single_point = np.array([10, 10, 10])
expected_point_single = np.array([[10, 0, 10]])
expected_distance_single = np.array([10])
multiple_point = np.array(
    [[10, 10, 10], [20, 10, 30], [20, 40, 20], [10, -5, 30]]
)
expected_multiple_point = np.array(
    [[10, 0, 10], [20, 0, 30], [20, 0, 20], [10, 0, 30]]
)
expected_distance_multiple = np.array([10, 10, 40, -5])


@pytest.mark.parametrize(
    "point,expected_projected_point,expected_distances",
    [
        (single_point, expected_point_single, expected_distance_single),
        (multiple_point, expected_multiple_point, expected_distance_multiple),
    ],
)
def test_project_point_to_plane(
    point, expected_projected_point, expected_distances
):
    plane_point = np.array([20, 0, 0])
    plane_normal = np.array([0, 1, 0])
    projected_point, distance_to_plane = project_points_onto_plane(
        point, plane_point, plane_normal
    )

    np.testing.assert_allclose(projected_point, expected_projected_point)
    np.testing.assert_allclose(distance_to_plane, expected_distances)


@pytest.mark.parametrize(
    "vec_1, vec_2",
    [
        (np.array([10, 0]), np.array([0, 5])),
        (np.array([0, 5]), np.array([0, 5])),
        (np.array([0, 5]), np.array([0, -5])),
    ],
)
def test_rotation_matrix_from_vectors_2d(vec_1, vec_2):

    rotation_matrix = rotation_matrix_from_vectors_2d(vec_1, vec_2)

    rotated_1 = rotation_matrix.dot(vec_1)
    unit_rotated_1 = rotated_1 / np.linalg.norm(rotated_1)

    unit_vec_2 = vec_2 / np.linalg.norm(vec_2)

    np.testing.assert_allclose(unit_rotated_1, unit_vec_2)


@pytest.mark.parametrize(
    "vec_1, vec_2",
    [
        (np.array([10, 0, 0]), np.array([0, 5, 0])),
        (np.array([0, 5, 0]), np.array([0, 5, 0])),
        (np.array([0, 5, 0]), np.array([0, -5, 0])),
    ],
)
def test_rotation_matrix_from_vectors_3d(vec_1, vec_2):
    """Test that calculated rotation matrices align vec1 to vec2."""
    rotation_matrix = rotation_matrix_from_vectors_3d(vec_1, vec_2)

    rotated_1 = rotation_matrix.dot(vec_1)
    unit_rotated_1 = rotated_1 / np.linalg.norm(rotated_1)

    unit_vec_2 = vec_2 / np.linalg.norm(vec_2)

    np.testing.assert_allclose(unit_rotated_1, unit_vec_2)


@pytest.mark.parametrize(
    "line_position, line_direction, plane_position, plane_normal, expected",
    [
        ([0, 0, 1], [0, 0, -1], [0, 0, 0], [0, 0, 1], [0, 0, 0]),
        ([1, 1, 1], [-1, -1, -1], [0, 0, 0], [0, 0, 1], [0, 0, 0]),
        ([2, 2, 2], [-1, -1, -1], [1, 1, 1], [0, 0, 1], [1, 1, 1]),
    ],
)
def test_intersect_line_with_plane_3d(
    line_position, line_direction, plane_position, plane_normal, expected
):
    """Test that arbitrary line-plane intersections are correctly calculated."""
    intersection = intersect_line_with_plane_3d(
        line_position, line_direction, plane_position, plane_normal
    )
    np.testing.assert_allclose(expected, intersection)


def test_intersect_line_with_multiple_planes_3d():
    """Test intersecting a ray with multiple planes and getting the intersection
    with each one.
    """
    line_position = [0, 0, 1]
    line_direction = [0, 0, -1]
    plane_positions = [[0, 0, 0], [0, 0, 1]]
    plane_normals = [[0, 0, 1], [0, 0, 1]]
    intersections = intersect_line_with_multiple_planes_3d(
        line_position, line_direction, plane_positions, plane_normals
    )

    expected = np.array([[0, 0, 0], [0, 0, 1]])
    np.testing.assert_allclose(intersections, expected)


@pytest.mark.parametrize(
    "point, bounding_box, expected",
    [
        ([5, 5, 5], np.array([[0, 10], [0, 10], [0, 10]]), [5, 5, 5]),
        ([10, 10, 10], np.array([[0, 10], [0, 10], [0, 10]]), [9, 9, 9]),
        ([5, 5, 15], np.array([[0, 10], [0, 10], [0, 10]]), [5, 5, 9]),
    ],
)
def test_clamp_point_to_bounding_box(point, bounding_box, expected):
    """Test that points are correctly clamped to the limits of the data.
    Note: bounding boxes are calculated from layer extents, points are clamped
    to the range of valid indices into each dimension.

    e.g. for a shape (10,) array, data is clamped to the range (0, 9)
    """
    clamped_point = clamp_point_to_bounding_box(point, bounding_box)
    np.testing.assert_allclose(expected, clamped_point)


def test_clamp_multiple_points_to_bounding_box():
    """test that an array of points can be clamped to the bbox"""
    points = np.array([[10, 10, 10], [0, 5, 0], [20, 0, 20]])
    bbox = np.array([[0, 25], [0, 10], [3, 25]])
    expected_points = np.array([[10, 9, 10], [0, 5, 3], [20, 0, 20]])
    clamped_points = clamp_point_to_bounding_box(points, bbox)
    np.testing.assert_array_equal(clamped_points, expected_points)


@pytest.mark.parametrize(
    'bounding_box, face_normal, expected',
    [
        (np.array([[5, 10], [10, 20], [20, 30]]), np.array([1, 0, 0]), 10),
        (np.array([[5, 10], [10, 20], [20, 30]]), np.array([-1, 0, 0]), 5),
        (np.array([[5, 10], [10, 20], [20, 30]]), np.array([0, 1, 0]), 20),
        (np.array([[5, 10], [10, 20], [20, 30]]), np.array([0, -1, 0]), 10),
        (np.array([[5, 10], [10, 20], [20, 30]]), np.array([0, 0, 1]), 30),
        (np.array([[5, 10], [10, 20], [20, 30]]), np.array([0, 0, -1]), 20),
    ],
)
def test_face_coordinate_from_bounding_box(
    bounding_box, face_normal, expected
):
    """Test that the correct face coordinate is calculated.

    Face coordinate is a float which is the value where a face of a bounding box,
    defined by a face normal, intersects the axis the normal vector is aligned with.
    """
    face_coordinate = face_coordinate_from_bounding_box(
        bounding_box, face_normal
    )
    np.testing.assert_allclose(expected, face_coordinate)


@pytest.mark.parametrize(
    'plane_intercept, plane_normal, line_start, line_direction, expected',
    [
        (
            0,
            np.array([0, 0, 1]),
            np.array([0, 0, 1]),
            np.array([0, 0, 1]),
            [0, 0, 0],
        ),
        (
            10,
            np.array([0, 0, 1]),
            np.array([0, 0, 0]),
            np.array([0, 0, 1]),
            [0, 0, 10],
        ),
        (
            10,
            np.array([0, 1, 0]),
            np.array([0, 1, 0]),
            np.array([0, 1, 0]),
            [0, 10, 0],
        ),
        (
            10,
            np.array([1, 0, 0]),
            np.array([1, 0, 0]),
            np.array([1, 0, 0]),
            [10, 0, 0],
        ),
    ],
)
def test_line_with_axis_aligned_plane(
    plane_intercept, plane_normal, line_start, line_direction, expected
):
    """Test that intersections between line and axis aligned plane are
    calculated correctly.
    """
    intersection = intersect_line_with_axis_aligned_plane(
        plane_intercept, plane_normal, line_start, line_direction
    )
    np.testing.assert_allclose(expected, intersection)


def test_bounding_box_to_face_vertices_3d():
    """Test that bounding_box_to_face_vertices returns a dictionary of vertices
    for each face of an axis aligned 3D bounding box.
    """
    bounding_box = np.array([[5, 10], [15, 20], [25, 30]])
    face_vertices = bounding_box_to_face_vertices(bounding_box)
    expected = {
        'x_pos': np.array(
            [[5, 15, 30], [5, 20, 30], [10, 20, 30], [10, 15, 30]]
        ),
        'x_neg': np.array(
            [[5, 15, 25], [5, 20, 25], [10, 20, 25], [10, 15, 25]]
        ),
        'y_pos': np.array(
            [[5, 20, 25], [5, 20, 30], [10, 20, 30], [10, 20, 25]]
        ),
        'y_neg': np.array(
            [[5, 15, 25], [5, 15, 30], [10, 15, 30], [10, 15, 25]]
        ),
        'z_pos': np.array(
            [[10, 15, 25], [10, 15, 30], [10, 20, 30], [10, 20, 25]]
        ),
        'z_neg': np.array(
            [[5, 15, 25], [5, 15, 30], [5, 20, 30], [5, 20, 25]]
        ),
    }
    for k in face_vertices:
        np.testing.assert_allclose(expected[k], face_vertices[k])


def test_bounding_box_to_face_vertices_nd():
    """Test that bounding_box_to_face_vertices returns a dictionary of vertices
    for each face of an axis aligned nD bounding box.
    """
    bounding_box = np.array([[0, 0], [0, 0], [5, 10], [15, 20], [25, 30]])
    face_vertices = bounding_box_to_face_vertices(bounding_box)
    expected = {
        'x_pos': np.array(
            [[5, 15, 30], [5, 20, 30], [10, 20, 30], [10, 15, 30]]
        ),
        'x_neg': np.array(
            [[5, 15, 25], [5, 20, 25], [10, 20, 25], [10, 15, 25]]
        ),
        'y_pos': np.array(
            [[5, 20, 25], [5, 20, 30], [10, 20, 30], [10, 20, 25]]
        ),
        'y_neg': np.array(
            [[5, 15, 25], [5, 15, 30], [10, 15, 30], [10, 15, 25]]
        ),
        'z_pos': np.array(
            [[10, 15, 25], [10, 15, 30], [10, 20, 30], [10, 20, 25]]
        ),
        'z_neg': np.array(
            [[5, 15, 25], [5, 15, 30], [5, 20, 30], [5, 20, 25]]
        ),
    }
    for k in face_vertices:
        np.testing.assert_allclose(expected[k], face_vertices[k])


@pytest.mark.parametrize(
    'triangle, expected',
    [
        (np.array([[[-1, -1], [-1, 1], [1, 0]]]), True),
        (np.array([[[1, 1], [2, 1], [1.5, 2]]]), False),
    ],
)
def test_inside_triangles(triangle, expected):
    """Test that inside triangles returns an array of True for triangles which
    contain the origin, False otherwise.
    """
    inside = np.all(inside_triangles(triangle))
    assert inside == expected


@pytest.mark.parametrize(
    'point, quadrilateral, expected',
    [
        (
            np.array([0.5, 0.5]),
            np.array([[0, 0], [0, 1], [1, 1], [0, 1]]),
            True,
        ),
        (np.array([2, 2]), np.array([[0, 0], [0, 1], [1, 0], [1, 1]]), False),
    ],
)
def test_point_in_quadrilateral_2d(point, quadrilateral, expected):
    """Test that point_in_quadrilateral_2d determines whether a point
    is inside a quadrilateral.
    """
    inside = point_in_quadrilateral_2d(point, quadrilateral)
    assert inside == expected


@pytest.mark.parametrize(
    'click_position, quadrilateral, view_dir, expected',
    [
        (
            np.array([0, 0, 0]),
            np.array([[-1, -1, 0], [-1, 1, 0], [1, 1, 0], [1, -1, 0]]),
            np.array([0, 0, 1]),
            True,
        ),
        (
            np.array([0, 0, 5]),
            np.array([[-1, -1, 0], [-1, 1, 0], [1, 1, 0], [1, -1, 0]]),
            np.array([0, 0, 1]),
            True,
        ),
        (
            np.array([0, 5, 0]),
            np.array([[-1, -1, 0], [-1, 1, 0], [1, 1, 0], [1, -1, 0]]),
            np.array([0, 0, 1]),
            False,
        ),
    ],
)
def test_click_in_quadrilateral_3d(
    click_position, quadrilateral, view_dir, expected
):
    """Test that click in quadrilateral 3d determines whether the projection
    of a 3D point onto a plane falls within a 3d quadrilateral projected
    onto the same plane
    """
    in_quadrilateral = line_in_quadrilateral_3d(
        click_position, view_dir, quadrilateral
    )
    assert in_quadrilateral == expected


@pytest.mark.parametrize(
    'click_position, bounding_box, view_dir, expected',
    [
        (
            np.array([5, 5, 5]),
            np.array([[0, 10], [0, 10], [0, 10]]),
            np.array([0, 0, 1]),
            ([0, 0, -1], [0, 0, 1]),
        ),
        (
            np.array([-5, -5, -5]),
            np.array([[0, 10], [0, 10], [0, 10]]),
            np.array([0, 0, 1]),
            (None, None),
        ),
        (
            np.array([5, 5, 5]),
            np.array([[0, 10], [0, 10], [0, 10]]),
            np.array([0, 1, 0]),
            ([0, -1, 0], [0, 1, 0]),
        ),
        (
            np.array([5, 5, 5]),
            np.array([[0, 10], [0, 10], [0, 10]]),
            np.array([1, 0, 0]),
            ([-1, 0, 0], [1, 0, 0]),
        ),
    ],
)
def test_find_front_back_face(
    click_position, bounding_box, view_dir, expected
):
    """Test that find front_back face finds the faces of an axis aligned
    bounding box that a ray intersects with.
    """
    result = find_front_back_face(click_position, bounding_box, view_dir)
    for idx, item in enumerate(result):
        if item is not None:
            np.testing.assert_allclose(item, expected[idx])
        else:
            assert item == expected[idx]


@pytest.mark.parametrize(
    'line_position, line_direction, bounding_box, face_normal, expected',
    [
        (
            np.array([5, 5, 5]),
            np.array([0, 0, 1]),
            np.array([[0, 10], [0, 10], [0, 10]]),
            np.array([0, 0, 1]),
            np.array([5, 5, 10]),
        ),
        (
            np.array([5, 5, 5]),
            np.array([0, 0, 1]),
            np.array([[0, 10], [0, 10], [0, 10]]),
            np.array([0, 0, -1]),
            np.array([5, 5, 0]),
        ),
        (
            np.array([5, 5, 5]),
            np.array([0, 1, 0]),
            np.array([[0, 10], [0, 10], [0, 10]]),
            np.array([0, 1, 0]),
            np.array([5, 10, 5]),
        ),
        (
            np.array([5, 5, 5]),
            np.array([0, 1, 0]),
            np.array([[0, 10], [0, 10], [0, 10]]),
            np.array([0, 1, 0]),
            np.array([5, 10, 5]),
        ),
        (
            np.array([5, 5, 5]),
            np.array([1, 0, 0]),
            np.array([[0, 10], [0, 10], [0, 10]]),
            np.array([1, 0, 0]),
            np.array([10, 5, 5]),
        ),
    ],
)
def test_intersect_line_with_axis_aligned_bounding_box_3d(
    line_position, line_direction, bounding_box, face_normal, expected
):
    """Test that intersections between lines and axis aligned
    bounding boxes are correctly computed.
    """
    result = intersect_line_with_axis_aligned_bounding_box_3d(
        line_position, line_direction, bounding_box, face_normal
    )
    np.testing.assert_allclose(expected, result)


def test_distance_between_point_and_line_3d():
    """Test that distance between points and lines are correctly computed."""
    line_position = np.random.random(size=3)
    line_direction = np.array([0, 0, 1])

    # find a point a random distance away on the line
    point_on_line = line_position + np.random.random(1) * line_direction

    # find a point a fixed distance from the point in a direction perpendicular
    # to the line direction.
    expected_distance = np.random.random(1)
    point = point_on_line + expected_distance * np.array([0, 1, 0])

    # calculate distance and check that it is correct
    distance = distance_between_point_and_line_3d(
        point, line_position, line_direction
    )

    np.testing.assert_allclose(distance, expected_distance)


def test_line_in_triangles_3d():
    line_point = np.array([0, 5, 5])
    line_direction = np.array([1, 0, 0])

    triangles = np.array(
        [
            [[10, 0, 0], [19, 10, 5], [5, 5, 10]],
            [[10, 4, 4], [10, 0, 0], [10, 4, 0]],
        ]
    )
    in_triangle = line_in_triangles_3d(line_point, line_direction, triangles)
    np.testing.assert_array_equal(in_triangle, [True, False])


@pytest.mark.parametrize(
    "ray_start,ray_direction,expected_index,expected_position",
    [
        ([0, 1, 1], [1, 0, 0], 0, [3, 1, 1]),
        ([6, 1, 1], [-1, 0, 0], 1, [5, 1, 1]),
    ],
)
def test_find_nearest_triangle_intersection(
    ray_start, ray_direction, expected_index, expected_position
):
    triangles = np.array(
        [
            [[3, 0, 0], [3, 0, 10], [3, 10, 0]],
            [[5, 0, 0], [5, 0, 10], [5, 10, 0]],
            [
                [2, 50, 50],
                [2, 50, 100],
                [2, 100, 50],
            ],
        ]
    )
    index, intersection = find_nearest_triangle_intersection(
        ray_position=ray_start,
        ray_direction=ray_direction,
        triangles=triangles,
    )

    assert index == expected_index
    np.testing.assert_allclose(intersection, expected_position)


def test_find_nearest_triangle_intersection_no_intersection():
    """Test find_nearest_triangle_intersection() when there is not intersection"""
    triangles = np.array(
        [
            [[3, 0, 0], [3, 0, 10], [3, 10, 0]],
            [[5, 0, 0], [5, 0, 10], [5, 10, 0]],
            [
                [2, 50, 50],
                [2, 50, 100],
                [2, 100, 50],
            ],
        ]
    )

    ray_start = np.array([0, -10, -10])
    ray_direction = np.array([-1, 0, 0])

    index, intersection = find_nearest_triangle_intersection(
        ray_position=ray_start,
        ray_direction=ray_direction,
        triangles=triangles,
    )

    assert index is None
    assert intersection is None



from enum import auto

from ..utils.misc import StringEnum


class LoopMode(StringEnum):
    """Looping mode for animating an axis.

    LoopMode.ONCE
        Animation will stop once movie reaches the max frame (if fps > 0) or
        the first frame (if fps < 0).
    LoopMode.LOOP
        Movie will return to the first frame after reaching the last frame,
        looping continuously until stopped.
    LoopMode.BACK_AND_FORTH
        Movie will loop continuously until stopped, reversing direction when
        the maximum or minimum frame has been reached.
    """

    ONCE = auto()
    LOOP = auto()
    BACK_AND_FORTH = auto()



from typing import Optional

from qtpy import QtCore
from qtpy.QtWidgets import (
    QApplication,
    QFrame,
    QHBoxLayout,
    QLabel,
    QProgressBar,
    QVBoxLayout,
    QWidget,
)

from ...utils.progress import progress


class QtLabeledProgressBar(QWidget):
    """QProgressBar with QLabels for description and ETA."""

    def __init__(
        self, parent: Optional[QWidget] = None, prog: progress = None
    ) -> None:
        super().__init__(parent)
        self.setAttribute(QtCore.Qt.WidgetAttribute.WA_DeleteOnClose)

        self.progress = prog

        self.qt_progress_bar = QProgressBar()
        self.description_label = QLabel()
        self.eta_label = QLabel()
        base_layout = QVBoxLayout()

        pbar_layout = QHBoxLayout()
        pbar_layout.addWidget(self.description_label)
        pbar_layout.addWidget(self.qt_progress_bar)
        pbar_layout.addWidget(self.eta_label)
        base_layout.addLayout(pbar_layout)

        line = QFrame(self)
        line.setObjectName("QtCustomTitleBarLine")
        line.setFixedHeight(1)
        base_layout.addWidget(line)

        self.setLayout(base_layout)

    def setRange(self, min, max):
        self.qt_progress_bar.setRange(min, max)

    def setValue(self, value):
        self.qt_progress_bar.setValue(value)
        QApplication.processEvents()

    def setDescription(self, value):
        if not value.endswith(': '):
            value = f'{value}: '
        self.description_label.setText(value)
        QApplication.processEvents()

    def _set_value(self, event):
        self.setValue(event.value)

    def _get_value(self):
        return self.qt_progress_bar.value()

    def _set_description(self, event):
        self.setDescription(event.value)

    def _make_indeterminate(self, event):
        self.setRange(0, 0)

    def _set_eta(self, event):
        self.eta_label.setText(event.value)

    def _set_total(self, event):
        self.setRange(0, event.value)


class QtProgressBarGroup(QWidget):
    """One or more QtLabeledProgressBars with a QFrame line separator at the bottom"""

    def __init__(
        self,
        qt_progress_bar: QtLabeledProgressBar,
        parent: Optional[QWidget] = None,
    ) -> None:
        super().__init__(parent)
        self.setAttribute(QtCore.Qt.WidgetAttribute.WA_DeleteOnClose)

        pbr_group_layout = QVBoxLayout()
        pbr_group_layout.addWidget(qt_progress_bar)
        pbr_group_layout.setContentsMargins(0, 0, 0, 0)

        line = QFrame(self)
        line.setObjectName("QtCustomTitleBarLine")
        line.setFixedHeight(1)
        pbr_group_layout.addWidget(line)

        self.setLayout(pbr_group_layout)



"""ImageSlice class.
"""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Callable

import numpy as np

from ...utils import config
from ._image_loader import ImageLoader
from ._image_slice_data import ImageSliceData
from ._image_view import ImageView

LOGGER = logging.getLogger("napari.loader")

if TYPE_CHECKING:
    from ...types import ArrayLike


def _create_loader_class() -> ImageLoader:
    """Return correct ImageLoader for sync or async.

    Returns
    -------
    ImageLoader
        Return ImageLoader for sync or ChunkImageLoader for async.
    """
    if config.async_loading:
        from .experimental._chunked_image_loader import ChunkedImageLoader

        return ChunkedImageLoader()
    else:
        return ImageLoader()


class ImageSlice:
    """The slice of the image that we are currently viewing.

    Parameters
    ----------
    image : ArrayLike
        The initial image used as the image and the thumbnail source.
    image_converter : Callable[[ArrayLike], ArrayLike]
        ImageView uses this to convert from raw to viewable.
    rgb : bool
        True if the image is RGB format. Otherwise its RGBA.

    Attributes
    ----------
    image : ImageView
        The main image for this slice.
    thumbnail : ImageView
        The source image used to compute the smaller thumbnail image.
    rgb : bool
        Is the image in RGB or RGBA format.
    loaded : bool
        Has the data for this slice been loaded yet.
    """

    def __init__(
        self,
        image: ArrayLike,
        image_converter: Callable[[ArrayLike], ArrayLike],
        rgb: bool = False,
    ):
        LOGGER.debug("ImageSlice.__init__")
        self.image: ImageView = ImageView(image, image_converter)
        self.thumbnail: ImageView = ImageView(image, image_converter)
        self.rgb = rgb
        self.loader = _create_loader_class()

        # With async there can be a gap between when the ImageSlice is
        # created and the data is actually loaded. However initialize
        # as True in case we aren't even doing async loading.
        self.loaded = True

    def _set_raw_images(
        self, image: ArrayLike, thumbnail_source: ArrayLike
    ) -> None:
        """Set the image and its thumbnail.

        If floating point / grayscale then clip to [0..1].

        Parameters
        ----------
        image : ArrayLike
            Set this as the main image.
        thumbnail_source : ArrayLike
            Derive the thumbnail from this image.
        """
        # Single scale images don't have a separate thumbnail so we just
        # use the image itself.
        if thumbnail_source is None:
            thumbnail_source = image

        if self.rgb and image.dtype.kind == 'f':
            image = np.clip(image, 0, 1)
            thumbnail_source = np.clip(thumbnail_source, 0, 1)
        self.image.raw = image

        # save a computation of view image if thumbnail and image is equal
        if thumbnail_source is image:
            self.thumbnail._raw = self.image._raw
            self.thumbnail._view = self.image._view
        else:
            self.thumbnail.raw = thumbnail_source

    def load(self, data: ImageSliceData) -> bool:
        """Load this data into the slice.

        Parameters
        ----------
        data : ImageSliceData
            The data to load into this slice.

        Returns
        -------
        bool
            Return True if load was synchronous.
        """
        self.loaded = False  # False until self._on_loaded is calls
        return self.loader.load(data)

    def on_loaded(self, data: ImageSliceData) -> bool:
        """Data was loaded, show the new data.

        Parameters
        ----------
        data : ImageSliceData
            The newly loaded data we want to show.

        Returns
        -------
        bool
            True if the data was used, False if was for the wrong slice.
        """
        if not self.loader.match(data):
            return False  # data was not used.

        # Display the newly loaded data.
        self._set_raw_images(data.image, data.thumbnail_source)
        self.loaded = True
        return True  # data was used.



import numpy as np

from napari.layers.points._points_utils import (
    _create_box_from_corners_3d,
    _points_in_box_3d,
)


def test_create_box_from_corners_3d():
    corners = np.array([[5, 0, 0], [5, 10, 10]])
    normal = np.array([1, 0, 0])
    up_dir = np.array([0, 1, 0])

    box = _create_box_from_corners_3d(
        box_corners=corners, box_normal=normal, up_vector=up_dir
    )

    expected_box = np.array([[5, 0, 0], [5, 0, 10], [5, 10, 10], [5, 10, 0]])
    np.testing.assert_allclose(box, expected_box)


def test_points_in_box_3d():
    normal = np.array([1, 0, 0])
    up_dir = np.array([0, 1, 0])
    corners = np.array([[10, 10, 10], [10, 20, 20]])
    points = np.array([[0, 15, 15], [10, 30, 25], [10, 12, 18], [20, 15, 30]])
    sizes = np.ones((points.shape[0],))

    inside = _points_in_box_3d(
        box_corners=corners,
        box_normal=normal,
        up_direction=up_dir,
        points=points,
        sizes=sizes,
    )

    assert set(inside) == {0, 2}



import os

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "not-installed"

# Allows us to use pydata/sparse arrays as layer data
os.environ.setdefault('SPARSE_AUTO_DENSIFY', '1')
del os

# Add everything that needs to be accessible from the napari namespace here.
_proto_all_ = [
    '__version__',
    'components',
    'experimental',
    'layers',
    'qt',
    'types',
    'viewer',
    'utils',
]

_submod_attrs = {
    '_event_loop': ['gui_qt', 'run'],
    'plugins.io': ['save_layers'],
    'utils': ['sys_info'],
    'utils.notifications': ['notification_manager'],
    'view_layers': [
        'view_image',
        'view_labels',
        'view_path',
        'view_points',
        'view_shapes',
        'view_surface',
        'view_tracks',
        'view_vectors',
        'imshow',
    ],
    'viewer': ['Viewer', 'current_viewer'],
}

# All imports in __init__ are hidden inside of `__getattr__` to prevent
# importing the full chain of packages required when calling `import napari`.
#
# This has the biggest implications for running `napari` on the command line
# (or running `python -m napari`) since `napari.__init__` gets imported
# on the way to `napari.__main__`. Importing everything here has the
# potential to take a second or more, so we definitely don't want to import it
# just to access the CLI (which may not actually need any of the imports)

from ._lazy import install_lazy

__getattr__, __dir__, __all__ = install_lazy(
    __name__, _proto_all_, _submod_attrs
)
del install_lazy



import numpy as np
import pandas as pd
import pytest

from napari.layers.utils.string_encoding import (
    ConstantStringEncoding,
    DirectStringEncoding,
    FormatStringEncoding,
    ManualStringEncoding,
    StringEncoding,
)


def make_features_with_no_columns(*, num_rows) -> pd.DataFrame:
    return pd.DataFrame({}, index=range(num_rows))


@pytest.fixture
def features() -> pd.DataFrame:
    return pd.DataFrame(
        {
            'class': ['a', 'b', 'c'],
            'confidence': [0.5, 1, 0.25],
        }
    )


@pytest.fixture
def numeric_features() -> pd.DataFrame:
    return pd.DataFrame(
        {
            'label': [1, 2, 3],
            'confidence': [0.5, 1, 0.25],
        }
    )


def test_constant_call_with_no_rows():
    features = make_features_with_no_columns(num_rows=0)
    encoding = ConstantStringEncoding(constant='abc')

    values = encoding(features)

    np.testing.assert_equal(values, 'abc')


def test_constant_call_with_some_rows():
    features = make_features_with_no_columns(num_rows=3)
    encoding = ConstantStringEncoding(constant='abc')

    values = encoding(features)

    np.testing.assert_equal(values, 'abc')


def test_manual_call_with_no_rows():
    features = make_features_with_no_columns(num_rows=0)
    array = ['a', 'b', 'c']
    default = 'd'
    encoding = ManualStringEncoding(array=array, default=default)

    values = encoding(features)

    np.testing.assert_array_equal(values, np.array([], dtype=str))


def test_manual_call_with_fewer_rows():
    features = make_features_with_no_columns(num_rows=2)
    array = ['a', 'b', 'c']
    default = 'd'
    encoding = ManualStringEncoding(array=array, default=default)

    values = encoding(features)

    np.testing.assert_array_equal(values, ['a', 'b'])


def test_manual_call_with_same_rows():
    features = make_features_with_no_columns(num_rows=3)
    array = ['a', 'b', 'c']
    default = 'd'
    encoding = ManualStringEncoding(array=array, default=default)

    values = encoding(features)

    np.testing.assert_array_equal(values, ['a', 'b', 'c'])


def test_manual_with_more_rows():
    features = make_features_with_no_columns(num_rows=4)
    array = ['a', 'b', 'c']
    default = 'd'
    encoding = ManualStringEncoding(array=array, default=default)

    values = encoding(features)

    np.testing.assert_array_equal(values, ['a', 'b', 'c', 'd'])


def test_direct(features):
    encoding = DirectStringEncoding(feature='class')
    values = encoding(features)
    np.testing.assert_array_equal(values, features['class'])


def test_direct_with_a_missing_feature(features):
    encoding = DirectStringEncoding(feature='not_class')
    with pytest.raises(KeyError):
        encoding(features)


def test_format(features):
    encoding = FormatStringEncoding(format='{class}: {confidence:.2f}')
    values = encoding(features)
    np.testing.assert_array_equal(values, ['a: 0.50', 'b: 1.00', 'c: 0.25'])


def test_format_with_bad_string(features):
    encoding = FormatStringEncoding(format='{class}: {confidence:.2f')
    with pytest.raises(ValueError):
        encoding(features)


def test_format_with_missing_field(features):
    encoding = FormatStringEncoding(format='{class}: {score:.2f}')
    with pytest.raises(KeyError):
        encoding(features)


def test_format_with_mixed_feature_numeric_types(numeric_features):
    encoding = FormatStringEncoding(format='{label:d}: {confidence:.2f}')
    values = encoding(numeric_features)
    np.testing.assert_array_equal(values, ['1: 0.50', '2: 1.00', '3: 0.25'])


def test_validate_from_format_string():
    argument = '{class}: {score:.2f}'
    expected = FormatStringEncoding(format=argument)

    actual = StringEncoding.validate(argument)

    assert actual == expected


def test_validate_from_non_format_string():
    argument = 'abc'
    expected = DirectStringEncoding(feature=argument)

    actual = StringEncoding.validate(argument)

    assert actual == expected


def test_validate_from_sequence():
    argument = ['a', 'b', 'c']
    expected = ManualStringEncoding(array=argument)

    actual = StringEncoding.validate(argument)

    assert actual == expected


def test_validate_from_constant_dict():
    constant = 'test'
    argument = {'constant': constant}
    expected = ConstantStringEncoding(constant=constant)

    actual = StringEncoding.validate(argument)

    assert actual == expected


def test_validate_from_manual_dict():
    array = ['a', 'b', 'c']
    default = 'd'
    argument = {'array': array, 'default': default}
    expected = ManualStringEncoding(array=array, default=default)

    actual = StringEncoding.validate(argument)

    assert actual == expected


def test_validate_from_direct_dict():
    feature = 'class'
    argument = {'feature': feature}
    expected = DirectStringEncoding(feature=feature)

    actual = StringEncoding.validate(argument)

    assert actual == expected


def test_validate_from_format_dict():
    format = '{class}: {score:.2f}'
    argument = {'format': format}
    expected = FormatStringEncoding(format=format)

    actual = StringEncoding.validate(argument)

    assert actual == expected



"""
These convenience functions will be useful for searching pypi for packages
that match the plugin naming convention, and retrieving related metadata.
"""
import json
from concurrent.futures import ThreadPoolExecutor
from functools import lru_cache
from typing import Dict, Iterator, List, Optional, Tuple, TypedDict, cast
from urllib.request import Request, urlopen

from npe2 import PackageMetadata

from .utils import normalized_name

PyPIname = str


@lru_cache
def _user_agent() -> str:
    """Return a user agent string for use in http requests."""
    import platform

    from napari import __version__

    from ..utils import misc

    if misc.running_as_bundled_app():
        env = 'briefcase'
    elif misc.running_as_constructor_app():
        env = 'constructor'
    elif misc.in_jupyter():
        env = 'jupyter'
    elif misc.in_ipython():
        env = 'ipython'
    else:
        env = 'python'

    parts = [
        ('napari', __version__),
        ('runtime', env),
        (platform.python_implementation(), platform.python_version()),
        (platform.system(), platform.release()),
    ]
    return ' '.join(f'{k}/{v}' for k, v in parts)


class SummaryDict(TypedDict):
    """Objects returned at https://npe2api.vercel.app/api/summary ."""

    name: PyPIname
    version: str
    display_name: str
    summary: str
    author: str
    license: str
    home_page: str


@lru_cache
def pypi_plugin_summaries() -> List[SummaryDict]:
    """Return PackageMetadata object for all known napari plugins."""
    url = "https://npe2api.vercel.app/api/summary"
    with urlopen(Request(url, headers={'User-Agent': _user_agent()})) as resp:
        return json.load(resp)


@lru_cache
def conda_map() -> Dict[PyPIname, Optional[str]]:
    """Return map of PyPI package name to conda_channel/package_name ()."""
    url = "https://npe2api.vercel.app/api/conda"
    with urlopen(Request(url, headers={'User-Agent': _user_agent()})) as resp:
        return json.load(resp)


def iter_napari_plugin_info() -> Iterator[Tuple[PackageMetadata, bool]]:
    """Iterator of tuples of ProjectInfo, Conda availability for all napari plugins."""
    with ThreadPoolExecutor() as executor:
        data = executor.submit(pypi_plugin_summaries)
        _conda = executor.submit(conda_map)

    conda = _conda.result()
    for info in data.result():
        _info = cast(Dict[str, str], dict(info))
        # TODO: use this better.
        # this would require changing the api that qt_plugin_dialog expects to
        # receive (and it doesn't currently receive this from the hub API)
        _info.pop("display_name", None)

        # TODO: I'd prefer we didn't normalize the name here, but it's needed for
        # parity with the hub api.  change this later.
        name = _info.pop("name")
        meta = PackageMetadata(name=normalized_name(name), **_info)
        yield meta, (name in conda)



from collections.abc import Iterable
from typing import Sequence, Union

import numpy as np

from ...utils.geometry import (
    inside_triangles,
    intersect_line_with_triangles,
    line_in_triangles_3d,
)
from ...utils.translations import trans
from ._mesh import Mesh
from ._shapes_constants import ShapeType, shape_classes
from ._shapes_models import Line, Path, Shape
from ._shapes_utils import triangles_intersect_box


class ShapeList:
    """List of shapes class.

    Parameters
    ----------
    data : list
        List of Shape objects
    ndisplay : int
        Number of displayed dimensions.

    Attributes
    ----------
    shapes : (N, ) list
        Shape objects.
    data : (N, ) list of (M, D) array
        Data arrays for each shape.
    ndisplay : int
        Number of displayed dimensions.
    slice_keys : (N, 2, P) array
        Array of slice keys for each shape. Each slice key has the min and max
        values of the P non-displayed dimensions, useful for slicing
        multidimensional shapes. If the both min and max values of shape are
        equal then the shape is entirely contained within the slice specified
        by those values.
    shape_types : (N, ) list of str
        Name of shape type for each shape.
    edge_color : (N x 4) np.ndarray
        Array of RGBA edge colors for each shape.
    face_color : (N x 4) np.ndarray
        Array of RGBA face colors for each shape.
    edge_widths : (N, ) list of float
        Edge width for each shape.
    z_indices : (N, ) list of int
        z-index for each shape.

    Notes
    -----
    _vertices : np.ndarray
        MxN array of all displayed vertices from all shapes where N is equal to ndisplay
    _index : np.ndarray
        Length M array with the index (0, ..., N-1) of each shape that each
        vertex corresponds to
    _z_index : np.ndarray
        Length N array with z_index of each shape
    _z_order : np.ndarray
        Length N array with z_order of each shape. This must be a permutation
        of (0, ..., N-1).
    _mesh : Mesh
        Mesh object containing all the mesh information that will ultimately
        be rendered.
    """

    def __init__(self, data=[], ndisplay=2):

        self._ndisplay = ndisplay
        self.shapes = []
        self._displayed = []
        self._slice_key = []
        self.displayed_vertices = []
        self.displayed_index = []
        self._vertices = np.empty((0, self.ndisplay))
        self._index = np.empty((0), dtype=int)
        self._z_index = np.empty((0), dtype=int)
        self._z_order = np.empty((0), dtype=int)

        self._mesh = Mesh(ndisplay=self.ndisplay)

        self._edge_color = np.empty((0, 4))
        self._face_color = np.empty((0, 4))

        for d in data:
            self.add(d)

    @property
    def data(self):
        """list of (M, D) array: data arrays for each shape."""
        return [s.data for s in self.shapes]

    @property
    def ndisplay(self):
        """int: Number of displayed dimensions."""
        return self._ndisplay

    @ndisplay.setter
    def ndisplay(self, ndisplay):
        if self.ndisplay == ndisplay:
            return

        self._ndisplay = ndisplay
        self._mesh.ndisplay = self.ndisplay
        self._vertices = np.empty((0, self.ndisplay))
        self._index = np.empty((0), dtype=int)
        for index in range(len(self.shapes)):
            shape = self.shapes[index]
            shape.ndisplay = self.ndisplay
            self.remove(index, renumber=False)
            self.add(shape, shape_index=index)
        self._update_z_order()

    @property
    def slice_keys(self):
        """(N, 2, P) array: slice key for each shape."""
        return np.array([s.slice_key for s in self.shapes])

    @property
    def shape_types(self):
        """list of str: shape types for each shape."""
        return [s.name for s in self.shapes]

    @property
    def edge_color(self):
        """(N x 4) np.ndarray: Array of RGBA edge colors for each shape"""
        return self._edge_color

    @edge_color.setter
    def edge_color(self, edge_color):
        self._set_color(edge_color, 'edge')

    @property
    def face_color(self):
        """(N x 4) np.ndarray: Array of RGBA face colors for each shape"""
        return self._face_color

    @face_color.setter
    def face_color(self, face_color):
        self._set_color(face_color, 'face')

    def _set_color(self, colors, attribute):
        """Set the face_color or edge_color property

        Parameters
        ----------
        colors : (N, 4) np.ndarray
            The value for setting edge or face_color. There must
            be one color for each shape
        attribute : str in {'edge', 'face'}
            The name of the attribute to set the color of.
            Should be 'edge' for edge_color or 'face' for face_color.
        """
        n_shapes = len(self.data)
        if not np.all(colors.shape == (n_shapes, 4)):
            raise ValueError(
                trans._(
                    '{attribute}_color must have shape ({n_shapes}, 4)',
                    deferred=True,
                    attribute=attribute,
                    n_shapes=n_shapes,
                )
            )

        update_method = getattr(self, f'update_{attribute}_colors')
        indices = np.arange(len(colors))
        update_method(indices, colors, update=False)
        self._update_displayed()

    @property
    def edge_widths(self):
        """list of float: edge width for each shape."""
        return [s.edge_width for s in self.shapes]

    @property
    def z_indices(self):
        """list of int: z-index for each shape."""
        return [s.z_index for s in self.shapes]

    @property
    def slice_key(self):
        """list: slice key for slicing n-dimensional shapes."""
        return self._slice_key

    @slice_key.setter
    def slice_key(self, slice_key):
        slice_key = list(slice_key)
        if not np.all(self._slice_key == slice_key):
            self._slice_key = slice_key
            self._update_displayed()

    def _update_displayed(self):
        """Update the displayed data based on the slice key."""
        # The list slice key is repeated to check against both the min and
        # max values stored in the shapes slice key.
        slice_key = np.array([self.slice_key, self.slice_key])

        # Slice key must exactly match mins and maxs of shape as then the
        # shape is entirely contained within the current slice.
        if len(self.shapes) > 0:
            self._displayed = np.all(self.slice_keys == slice_key, axis=(1, 2))
        else:
            self._displayed = []
        disp_indices = np.where(self._displayed)[0]

        z_order = self._mesh.triangles_z_order
        disp_tri = np.isin(
            self._mesh.triangles_index[z_order, 0], disp_indices
        )
        self._mesh.displayed_triangles = self._mesh.triangles[z_order][
            disp_tri
        ]
        self._mesh.displayed_triangles_index = self._mesh.triangles_index[
            z_order
        ][disp_tri]
        self._mesh.displayed_triangles_colors = self._mesh.triangles_colors[
            z_order
        ][disp_tri]

        disp_vert = np.isin(self._index, disp_indices)
        self.displayed_vertices = self._vertices[disp_vert]
        self.displayed_index = self._index[disp_vert]

    def add(
        self,
        shape: Union[Shape, Sequence[Shape]],
        face_color=None,
        edge_color=None,
        shape_index=None,
        z_refresh=True,
    ):
        """Adds a single Shape object (single add mode) or multiple Shapes (multiple shape mode, which is much faster)

        If shape is a single instance of subclass Shape then single add mode will be used, otherwise multiple add mode

        Parameters
        ----------
        shape : single Shape or iterable of Shape
            Each shape must be a subclass of Shape, one of "{'Line', 'Rectangle',
            'Ellipse', 'Path', 'Polygon'}"
        face_color : color (or iterable of colors of same length as shape)
        edge_color : color (or iterable of colors of same length as shape)
        shape_index : None | int
            If int then edits the shape date at current index. To be used in
            conjunction with `remove` when renumber is `False`. If None, then
            appends a new shape to end of shapes list
            Must be None in multiple shape mode.
        z_refresh : bool
            If set to true, the mesh elements are reindexed with the new z order.
            When shape_index is provided, z_refresh will be overwritten to false,
            as the z indices will not change.
            When adding a batch of shapes, set to false  and then call
            ShapesList._update_z_order() once at the end.
        """
        # single shape mode
        if issubclass(type(shape), Shape):
            self._add_single_shape(
                shape=shape,
                face_color=face_color,
                edge_color=edge_color,
                shape_index=shape_index,
                z_refresh=z_refresh,
            )
        # multiple shape mode
        elif isinstance(shape, Iterable):
            if shape_index is not None:
                raise ValueError(
                    trans._(
                        'shape_index must be None when adding multiple shapes',
                        deferred=True,
                    )
                )
            self._add_multiple_shapes(
                shapes=shape,
                face_colors=face_color,
                edge_colors=edge_color,
                z_refresh=z_refresh,
            )
        else:
            raise ValueError(
                trans._(
                    'Cannot add single nor multiple shape',
                    deferred=True,
                )
            )

    def _add_single_shape(
        self,
        shape,
        face_color=None,
        edge_color=None,
        shape_index=None,
        z_refresh=True,
    ):
        """Adds a single Shape object

        Parameters
        ----------
        shape : subclass Shape
            Must be a subclass of Shape, one of "{'Line', 'Rectangle',
            'Ellipse', 'Path', 'Polygon'}"
        shape_index : None | int
            If int then edits the shape date at current index. To be used in
            conjunction with `remove` when renumber is `False`. If None, then
            appends a new shape to end of shapes list
        z_refresh : bool
            If set to true, the mesh elements are reindexed with the new z order.
            When shape_index is provided, z_refresh will be overwritten to false,
            as the z indices will not change.
            When adding a batch of shapes, set to false  and then call
            ShapesList._update_z_order() once at the end.
        """
        if not issubclass(type(shape), Shape):
            raise ValueError(
                trans._(
                    'shape must be subclass of Shape',
                    deferred=True,
                )
            )

        if shape_index is None:
            shape_index = len(self.shapes)
            self.shapes.append(shape)
            self._z_index = np.append(self._z_index, shape.z_index)

            if face_color is None:
                face_color = np.array([1, 1, 1, 1])
            self._face_color = np.vstack([self._face_color, face_color])
            if edge_color is None:
                edge_color = np.array([0, 0, 0, 1])
            self._edge_color = np.vstack([self._edge_color, edge_color])
        else:
            z_refresh = False
            self.shapes[shape_index] = shape
            self._z_index[shape_index] = shape.z_index

            if face_color is None:
                face_color = self._face_color[shape_index]
            else:
                self._face_color[shape_index, :] = face_color
            if edge_color is None:
                edge_color = self._edge_color[shape_index]
            else:
                self._edge_color[shape_index, :] = edge_color

        self._vertices = np.append(
            self._vertices, shape.data_displayed, axis=0
        )
        index = np.repeat(shape_index, len(shape.data))
        self._index = np.append(self._index, index, axis=0)

        # Add faces to mesh
        m = len(self._mesh.vertices)
        vertices = shape._face_vertices
        self._mesh.vertices = np.append(self._mesh.vertices, vertices, axis=0)
        vertices = shape._face_vertices
        self._mesh.vertices_centers = np.append(
            self._mesh.vertices_centers, vertices, axis=0
        )
        vertices = np.zeros(shape._face_vertices.shape)
        self._mesh.vertices_offsets = np.append(
            self._mesh.vertices_offsets, vertices, axis=0
        )
        index = np.repeat([[shape_index, 0]], len(vertices), axis=0)
        self._mesh.vertices_index = np.append(
            self._mesh.vertices_index, index, axis=0
        )

        triangles = shape._face_triangles + m
        self._mesh.triangles = np.append(
            self._mesh.triangles, triangles, axis=0
        )
        index = np.repeat([[shape_index, 0]], len(triangles), axis=0)
        self._mesh.triangles_index = np.append(
            self._mesh.triangles_index, index, axis=0
        )
        color_array = np.repeat([face_color], len(triangles), axis=0)
        self._mesh.triangles_colors = np.append(
            self._mesh.triangles_colors, color_array, axis=0
        )

        # Add edges to mesh
        m = len(self._mesh.vertices)
        vertices = (
            shape._edge_vertices + shape.edge_width * shape._edge_offsets
        )
        self._mesh.vertices = np.append(self._mesh.vertices, vertices, axis=0)
        vertices = shape._edge_vertices
        self._mesh.vertices_centers = np.append(
            self._mesh.vertices_centers, vertices, axis=0
        )
        vertices = shape._edge_offsets
        self._mesh.vertices_offsets = np.append(
            self._mesh.vertices_offsets, vertices, axis=0
        )
        index = np.repeat([[shape_index, 1]], len(vertices), axis=0)
        self._mesh.vertices_index = np.append(
            self._mesh.vertices_index, index, axis=0
        )

        triangles = shape._edge_triangles + m
        self._mesh.triangles = np.append(
            self._mesh.triangles, triangles, axis=0
        )
        index = np.repeat([[shape_index, 1]], len(triangles), axis=0)
        self._mesh.triangles_index = np.append(
            self._mesh.triangles_index, index, axis=0
        )
        color_array = np.repeat([edge_color], len(triangles), axis=0)
        self._mesh.triangles_colors = np.append(
            self._mesh.triangles_colors, color_array, axis=0
        )

        if z_refresh:
            # Set z_order
            self._update_z_order()

    def _add_multiple_shapes(
        self,
        shapes,
        face_colors=None,
        edge_colors=None,
        z_refresh=True,
    ):
        """Add multiple shapes at once (faster than adding them one by one)

        Parameters
        ----------
        shapes : iterable of Shape
            Each Shape must be a subclass of Shape, one of "{'Line', 'Rectangle',
            'Ellipse', 'Path', 'Polygon'}"
        face_colors : iterable of face_color
        edge_colors : iterable of edge_color
        z_refresh : bool
            If set to true, the mesh elements are reindexed with the new z order.
            When shape_index is provided, z_refresh will be overwritten to false,
            as the z indices will not change.
            When adding a batch of shapes, set to false  and then call
            ShapesList._update_z_order() once at the end.

        TODO: Currently shares a lot of code with `add()`, with the
        difference being that `add()` supports inserting shapes at a specific
        `shape_index`, whereas `add_multiple` will append them as a full batch
        """

        def _make_index(length, shape_index, cval=0):
            """Same but faster than `np.repeat([[shape_index, cval]], length, axis=0)`"""
            index = np.empty((length, 2), np.int32)
            index.fill(cval)
            index[:, 0] = shape_index
            return index

        all_z_index = []
        all_vertices = []
        all_index = []
        all_mesh_vertices = []
        all_mesh_vertices_centers = []
        all_mesh_vertices_offsets = []
        all_mesh_vertices_index = []
        all_mesh_triangles = []
        all_mesh_triangles_index = []
        all_mesh_triangles_colors = []

        m_mesh_vertices_count = len(self._mesh.vertices)

        if face_colors is None:
            face_colors = np.tile(np.array([1, 1, 1, 1]), (len(shapes), 1))
        else:
            face_colors = np.asarray(face_colors)

        if edge_colors is None:
            edge_colors = np.tile(np.array([0, 0, 0, 1]), (len(shapes), 1))
        else:
            edge_colors = np.asarray(edge_colors)

        if not len(face_colors) == len(edge_colors) == len(shapes):
            raise ValueError(
                trans._(
                    'shapes, face_colors, and edge_colors must be the same length',
                    deferred=True,
                )
            )

        if not all(issubclass(type(shape), Shape) for shape in shapes):
            raise ValueError(
                trans._(
                    'all shapes must be subclass of Shape',
                    deferred=True,
                )
            )

        for shape, face_color, edge_color in zip(
            shapes, face_colors, edge_colors
        ):

            shape_index = len(self.shapes)
            self.shapes.append(shape)
            all_z_index.append(shape.z_index)
            all_vertices.append(shape.data_displayed)
            all_index.append([shape_index] * len(shape.data))

            # Add faces to mesh
            m_tmp = m_mesh_vertices_count
            all_mesh_vertices.append(shape._face_vertices)
            m_mesh_vertices_count += len(shape._face_vertices)
            all_mesh_vertices_centers.append(shape._face_vertices)
            vertices = np.zeros(shape._face_vertices.shape)
            all_mesh_vertices_offsets.append(vertices)
            all_mesh_vertices_index.append(
                _make_index(len(vertices), shape_index, cval=0)
            )

            triangles = shape._face_triangles + m_tmp
            all_mesh_triangles.append(triangles)
            all_mesh_triangles_index.append(
                _make_index(len(triangles), shape_index, cval=0)
            )

            color_array = np.repeat([face_color], len(triangles), axis=0)
            all_mesh_triangles_colors.append(color_array)

            # Add edges to mesh
            m_tmp = m_mesh_vertices_count

            vertices = (
                shape._edge_vertices + shape.edge_width * shape._edge_offsets
            )
            all_mesh_vertices.append(vertices)
            m_mesh_vertices_count += len(vertices)

            all_mesh_vertices_centers.append(shape._edge_vertices)

            all_mesh_vertices_offsets.append(shape._edge_offsets)

            all_mesh_vertices_index.append(
                _make_index(len(shape._edge_offsets), shape_index, cval=1)
            )

            triangles = shape._edge_triangles + m_tmp
            all_mesh_triangles.append(triangles)
            all_mesh_triangles_index.append(
                _make_index(len(triangles), shape_index, cval=1)
            )

            color_array = np.repeat([edge_color], len(triangles), axis=0)
            all_mesh_triangles_colors.append(color_array)

        # assemble properties
        self._z_index = np.append(self._z_index, np.array(all_z_index), axis=0)
        self._face_color = np.vstack((self._face_color, face_colors))
        self._edge_color = np.vstack((self._edge_color, edge_colors))
        self._vertices = np.vstack((self._vertices, np.vstack(all_vertices)))
        self._index = np.append(self._index, np.concatenate(all_index), axis=0)

        self._mesh.vertices = np.vstack(
            (self._mesh.vertices, np.vstack(all_mesh_vertices))
        )
        self._mesh.vertices_centers = np.vstack(
            (self._mesh.vertices_centers, np.vstack(all_mesh_vertices_centers))
        )
        self._mesh.vertices_offsets = np.vstack(
            (self._mesh.vertices_offsets, np.vstack(all_mesh_vertices_offsets))
        )
        self._mesh.vertices_index = np.vstack(
            (self._mesh.vertices_index, np.vstack(all_mesh_vertices_index))
        )

        self._mesh.triangles = np.vstack(
            (self._mesh.triangles, np.vstack(all_mesh_triangles))
        )
        self._mesh.triangles_index = np.vstack(
            (self._mesh.triangles_index, np.vstack(all_mesh_triangles_index))
        )
        self._mesh.triangles_colors = np.vstack(
            (self._mesh.triangles_colors, np.vstack(all_mesh_triangles_colors))
        )

        if z_refresh:
            # Set z_order
            self._update_z_order()

    def remove_all(self):
        """Removes all shapes"""
        self.shapes = []
        self._vertices = np.empty((0, self.ndisplay))
        self._index = np.empty((0), dtype=int)
        self._z_index = np.empty((0), dtype=int)
        self._z_order = np.empty((0), dtype=int)
        self._mesh.clear()
        self._update_displayed()

    def remove(self, index, renumber=True):
        """Removes a single shape located at index.

        Parameters
        ----------
        index : int
            Location in list of the shape to be removed.
        renumber : bool
            Bool to indicate whether to renumber all shapes or not. If not the
            expectation is that this shape is being immediately added back to the
            list using `add_shape`.
        """
        indices = self._index != index
        self._vertices = self._vertices[indices]
        self._index = self._index[indices]

        # Remove triangles
        indices = self._mesh.triangles_index[:, 0] != index
        self._mesh.triangles = self._mesh.triangles[indices]
        self._mesh.triangles_colors = self._mesh.triangles_colors[indices]
        self._mesh.triangles_index = self._mesh.triangles_index[indices]

        # Remove vertices
        indices = self._mesh.vertices_index[:, 0] != index
        self._mesh.vertices = self._mesh.vertices[indices]
        self._mesh.vertices_centers = self._mesh.vertices_centers[indices]
        self._mesh.vertices_offsets = self._mesh.vertices_offsets[indices]
        self._mesh.vertices_index = self._mesh.vertices_index[indices]
        indices = np.where(np.invert(indices))[0]
        num_indices = len(indices)
        if num_indices > 0:
            indices = self._mesh.triangles > indices[0]
            self._mesh.triangles[indices] = (
                self._mesh.triangles[indices] - num_indices
            )

        if renumber:
            del self.shapes[index]
            indices = self._index > index
            self._index[indices] = self._index[indices] - 1
            self._z_index = np.delete(self._z_index, index)
            indices = self._mesh.triangles_index[:, 0] > index
            self._mesh.triangles_index[indices, 0] = (
                self._mesh.triangles_index[indices, 0] - 1
            )
            indices = self._mesh.vertices_index[:, 0] > index
            self._mesh.vertices_index[indices, 0] = (
                self._mesh.vertices_index[indices, 0] - 1
            )
            self._update_z_order()

    def _update_mesh_vertices(self, index, edge=False, face=False):
        """Updates the mesh vertex data and vertex data for a single shape
        located at index.

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        edge : bool
            Bool to indicate whether to update mesh vertices corresponding to
            edges
        face : bool
            Bool to indicate whether to update mesh vertices corresponding to
            faces and to update the underlying shape vertices
        """
        shape = self.shapes[index]
        if edge:
            indices = np.all(self._mesh.vertices_index == [index, 1], axis=1)
            self._mesh.vertices[indices] = (
                shape._edge_vertices + shape.edge_width * shape._edge_offsets
            )
            self._mesh.vertices_centers[indices] = shape._edge_vertices
            self._mesh.vertices_offsets[indices] = shape._edge_offsets
            self._update_displayed()

        if face:
            indices = np.all(self._mesh.vertices_index == [index, 0], axis=1)
            self._mesh.vertices[indices] = shape._face_vertices
            self._mesh.vertices_centers[indices] = shape._face_vertices
            indices = self._index == index
            self._vertices[indices] = shape.data_displayed
            self._update_displayed()

    def _update_z_order(self):
        """Updates the z order of the triangles given the z_index list"""
        self._z_order = np.argsort(self._z_index)
        if len(self._z_order) == 0:
            self._mesh.triangles_z_order = np.empty((0), dtype=int)
        else:
            _, idx, counts = np.unique(
                self._mesh.triangles_index[:, 0],
                return_index=True,
                return_counts=True,
            )
            triangles_z_order = [
                np.arange(idx[z], idx[z] + counts[z]) for z in self._z_order
            ]
            self._mesh.triangles_z_order = np.concatenate(triangles_z_order)
        self._update_displayed()

    def edit(
        self, index, data, face_color=None, edge_color=None, new_type=None
    ):
        """Updates the data of a single shape located at index. If
        `new_type` is not None then converts the shape type to the new type

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        data : np.ndarray
            NxD array of vertices.
        new_type : None | str | Shape
            If string , must be one of "{'line', 'rectangle', 'ellipse',
            'path', 'polygon'}".
        """
        if new_type is not None:
            cur_shape = self.shapes[index]
            if type(new_type) == str:
                shape_type = ShapeType(new_type)
                if shape_type in shape_classes.keys():
                    shape_cls = shape_classes[shape_type]
                else:
                    raise ValueError(
                        trans._(
                            '{shape_type} must be one of {shape_classes}',
                            deferred=True,
                            shape_type=shape_type,
                            shape_classes=set(shape_classes),
                        )
                    )
            else:
                shape_cls = new_type
            shape = shape_cls(
                data,
                edge_width=cur_shape.edge_width,
                z_index=cur_shape.z_index,
                dims_order=cur_shape.dims_order,
            )
        else:
            shape = self.shapes[index]
            shape.data = data

        if face_color is not None:
            self._face_color[index] = face_color
        if edge_color is not None:
            self._edge_color[index] = edge_color

        self.remove(index, renumber=False)
        self.add(shape, shape_index=index)
        self._update_z_order()

    def update_edge_width(self, index, edge_width):
        """Updates the edge width of a single shape located at index.

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        edge_width : float
            thickness of lines and edges.
        """
        self.shapes[index].edge_width = edge_width
        self._update_mesh_vertices(index, edge=True)

    def update_edge_color(self, index, edge_color, update=True):
        """Updates the edge color of a single shape located at index.

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        edge_color : str | tuple
            If string can be any color name recognized by vispy or hex value if
            starting with `#`. If array-like must be 1-dimensional array with 3
            or 4 elements.
        update : bool
            If True, update the mesh with the new color property. Set to False to avoid
            repeated updates when modifying multiple shapes. Default is True.
        """
        self._edge_color[index] = edge_color
        indices = np.all(self._mesh.triangles_index == [index, 1], axis=1)
        self._mesh.triangles_colors[indices] = self._edge_color[index]
        if update:
            self._update_displayed()

    def update_edge_colors(self, indices, edge_colors, update=True):
        """same as update_edge_color() but for multiple indices/edgecolors at once"""
        self._edge_color[indices] = edge_colors
        all_indices = np.bitwise_and(
            np.isin(self._mesh.triangles_index[:, 0], indices),
            self._mesh.triangles_index[:, 1] == 1,
        )
        self._mesh.triangles_colors[all_indices] = self._edge_color[
            self._mesh.triangles_index[all_indices, 0]
        ]
        if update:
            self._update_displayed()

    def update_face_color(self, index, face_color, update=True):
        """Updates the face color of a single shape located at index.

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        face_color : str | tuple
            If string can be any color name recognized by vispy or hex value if
            starting with `#`. If array-like must be 1-dimensional array with 3
            or 4 elements.
        update : bool
            If True, update the mesh with the new color property. Set to False to avoid
            repeated updates when modifying multiple shapes. Default is True.
        """
        self._face_color[index] = face_color
        indices = np.all(self._mesh.triangles_index == [index, 0], axis=1)
        self._mesh.triangles_colors[indices] = self._face_color[index]
        if update:
            self._update_displayed()

    def update_face_colors(self, indices, face_colors, update=True):
        """same as update_face_color() but for multiple indices/facecolors at once"""
        self._face_color[indices] = face_colors
        all_indices = np.bitwise_and(
            np.isin(self._mesh.triangles_index[:, 0], indices),
            self._mesh.triangles_index[:, 1] == 0,
        )
        self._mesh.triangles_colors[all_indices] = self._face_color[
            self._mesh.triangles_index[all_indices, 0]
        ]
        if update:
            self._update_displayed()

    def update_dims_order(self, dims_order):
        """Updates dimensions order for all shapes.

        Parameters
        ----------
        dims_order : (D,) list
            Order that the dimensions are rendered in.
        """
        for index in range(len(self.shapes)):
            if not self.shapes[index].dims_order == dims_order:
                shape = self.shapes[index]
                shape.dims_order = dims_order
                self.remove(index, renumber=False)
                self.add(shape, shape_index=index)
        self._update_z_order()

    def update_z_index(self, index, z_index):
        """Updates the z order of a single shape located at index.

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        z_index : int
            Specifier of z order priority. Shapes with higher z order are
            displayed ontop of others.
        """
        self.shapes[index].z_index = z_index
        self._z_index[index] = z_index
        self._update_z_order()

    def shift(self, index, shift):
        """Performs a 2D shift on a single shape located at index

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        shift : np.ndarray
            length 2 array specifying shift of shapes.
        """
        self.shapes[index].shift(shift)
        self._update_mesh_vertices(index, edge=True, face=True)

    def scale(self, index, scale, center=None):
        """Performs a scaling on a single shape located at index

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        scale : float, list
            scalar or list specifying rescaling of shape.
        center : list
            length 2 list specifying coordinate of center of scaling.
        """
        self.shapes[index].scale(scale, center=center)
        shape = self.shapes[index]
        self.remove(index, renumber=False)
        self.add(shape, shape_index=index)
        self._update_z_order()

    def rotate(self, index, angle, center=None):
        """Performs a rotation on a single shape located at index

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        angle : float
            angle specifying rotation of shape in degrees.
        center : list
            length 2 list specifying coordinate of center of rotation.
        """
        self.shapes[index].rotate(angle, center=center)
        self._update_mesh_vertices(index, edge=True, face=True)

    def flip(self, index, axis, center=None):
        """Performs an vertical flip on a single shape located at index

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        axis : int
            integer specifying axis of flip. `0` flips horizontal, `1` flips
            vertical.
        center : list
            length 2 list specifying coordinate of center of flip axes.
        """
        self.shapes[index].flip(axis, center=center)
        self._update_mesh_vertices(index, edge=True, face=True)

    def transform(self, index, transform):
        """Performs a linear transform on a single shape located at index

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        transform : np.ndarray
            2x2 array specifying linear transform.
        """
        self.shapes[index].transform(transform)
        shape = self.shapes[index]
        self.remove(index, renumber=False)
        self.add(shape, shape_index=index)
        self._update_z_order()

    def outline(self, indices):
        """Finds outlines of shapes listed in indices

        Parameters
        ----------
        indices : int | list
            Location in list of the shapes to be outline. If list must be a
            list of int

        Returns
        -------
        centers : np.ndarray
            Nx2 array of centers of outline
        offsets : np.ndarray
            Nx2 array of offsets of outline
        triangles : np.ndarray
            Mx3 array of any indices of vertices for triangles of outline
        """
        if type(indices) is list:
            meshes = self._mesh.triangles_index
            triangle_indices = [
                i
                for i, x in enumerate(meshes)
                if x[0] in indices and x[1] == 1
            ]
            meshes = self._mesh.vertices_index
            vertices_indices = [
                i
                for i, x in enumerate(meshes)
                if x[0] in indices and x[1] == 1
            ]
        else:
            triangle_indices = np.all(
                self._mesh.triangles_index == [indices, 1], axis=1
            )
            triangle_indices = np.where(triangle_indices)[0]
            vertices_indices = np.all(
                self._mesh.vertices_index == [indices, 1], axis=1
            )
            vertices_indices = np.where(vertices_indices)[0]

        offsets = self._mesh.vertices_offsets[vertices_indices]
        centers = self._mesh.vertices_centers[vertices_indices]
        triangles = self._mesh.triangles[triangle_indices]

        if type(indices) is list:
            t_ind = self._mesh.triangles_index[triangle_indices][:, 0]
            inds = self._mesh.vertices_index[vertices_indices][:, 0]
            starts = np.unique(inds, return_index=True)[1]
            for i, ind in enumerate(indices):
                inds = t_ind == ind
                adjust_index = starts[i] - vertices_indices[starts[i]]
                triangles[inds] = triangles[inds] + adjust_index
        else:
            triangles = triangles - vertices_indices[0]

        return centers, offsets, triangles

    def shapes_in_box(self, corners):
        """Determines which shapes, if any, are inside an axis aligned box.

        Looks only at displayed shapes

        Parameters
        ----------
        corners : np.ndarray
            2x2 array of two corners that will be used to create an axis
            aligned box.

        Returns
        -------
        shapes : list
            List of shapes that are inside the box.
        """

        triangles = self._mesh.vertices[self._mesh.displayed_triangles]
        intersects = triangles_intersect_box(triangles, corners)
        shapes = self._mesh.displayed_triangles_index[intersects, 0]
        shapes = np.unique(shapes).tolist()

        return shapes

    def inside(self, coord):
        """Determines if any shape at given coord by looking inside triangle
        meshes. Looks only at displayed shapes

        Parameters
        ----------
        coord : sequence of float
            Image coordinates to check if any shapes are at.

        Returns
        -------
        shape : int | None
            Index of shape if any that is at the coordinates. Returns `None`
            if no shape is found.
        """
        triangles = self._mesh.vertices[self._mesh.displayed_triangles]
        indices = inside_triangles(triangles - coord)
        shapes = self._mesh.displayed_triangles_index[indices, 0]

        if len(shapes) > 0:
            z_list = self._z_order.tolist()
            order_indices = np.array([z_list.index(m) for m in shapes])
            ordered_shapes = shapes[np.argsort(order_indices)]
            return ordered_shapes[0]
        else:
            return None

    def _inside_3d(self, ray_position: np.ndarray, ray_direction: np.ndarray):
        """Determines if any shape is intersected by a ray by looking inside triangle
        meshes. Looks only at displayed shapes.

        Parameters
        ----------
        ray_position : np.ndarray
            (3,) array containing the location that was clicked. This
            should be in the same coordinate system as the vertices.
        ray_direction : np.ndarray
            (3,) array describing the direction camera is pointing in
            the scene. This should be in the same coordinate system as
            the vertices.

        Returns
        -------
        shape : int | None
            Index of shape if any that is at the coordinates. Returns `None`
            if no shape is found.
        intersection_point : Optional[np.ndarray]
            The point where the ray intersects the mesh face. If there was
            no intersection, returns None.
        """
        triangles = self._mesh.vertices[self._mesh.displayed_triangles]
        inside = line_in_triangles_3d(
            line_point=ray_position,
            line_direction=ray_direction,
            triangles=triangles,
        )
        intersected_shapes = self._mesh.displayed_triangles_index[inside, 0]
        if len(intersected_shapes) > 0:
            intersection_points = self._triangle_intersection(
                triangle_indices=inside,
                ray_position=ray_position,
                ray_direction=ray_direction,
            )
            start_to_intersection = intersection_points - ray_position
            distances = np.linalg.norm(start_to_intersection, axis=1)
            closest_shape_index = np.argmin(distances)
            shape = intersected_shapes[closest_shape_index]
            intersection = intersection_points[closest_shape_index]
            return shape, intersection
        else:
            return None, None

    def _triangle_intersection(
        self,
        triangle_indices: np.ndarray,
        ray_position: np.ndarray,
        ray_direction: np.ndarray,
    ):
        """Find the intersection of a ray with specified triangles.

        Parameters
        ----------
        triangle_indices : np.ndarray
            (n,) array of shape indices to find the intersection with the ray. The indices should
            correspond with self._mesh.displayed_triangles.
        ray_position : np.ndarray
            (3,) array with the coordinate of the starting point of the ray in layer coordinates.
            Only provide the 3 displayed dimensions.
        ray_direction : np.ndarray
            (3,) array of the normal direction of the ray in layer coordinates.
            Only provide the 3 displayed dimensions.

        Returns
        -------
        intersection_points : np.ndarray
            (n x 3) array of the intersection of the ray with each of the specified shapes in layer coordinates.
            Only the 3 displayed dimensions are provided.
        """
        triangles = self._mesh.vertices[self._mesh.displayed_triangles]
        intersected_triangles = triangles[triangle_indices]
        intersection_points = intersect_line_with_triangles(
            line_point=ray_position,
            line_direction=ray_direction,
            triangles=intersected_triangles,
        )
        return intersection_points

    def to_masks(self, mask_shape=None, zoom_factor=1, offset=[0, 0]):
        """Returns N binary masks, one for each shape, embedded in an array of
        shape `mask_shape`.

        Parameters
        ----------
        mask_shape : np.ndarray | tuple | None
            2-tuple defining shape of mask to be generated. If non specified,
            takes the max of all the vertices
        zoom_factor : float
            Premultiplier applied to coordinates before generating mask. Used
            for generating as downsampled mask.
        offset : 2-tuple
            Offset subtracted from coordinates before multiplying by the
            zoom_factor. Used for putting negative coordinates into the mask.

        Returns
        -------
        masks : (N, M, P) np.ndarray
            Array where there is one binary mask of shape MxP for each of
            N shapes
        """
        if mask_shape is None:
            mask_shape = self.displayed_vertices.max(axis=0).astype('int')

        masks = np.array(
            [
                s.to_mask(mask_shape, zoom_factor=zoom_factor, offset=offset)
                for s in self.shapes
            ]
        )

        return masks

    def to_labels(self, labels_shape=None, zoom_factor=1, offset=[0, 0]):
        """Returns a integer labels image, where each shape is embedded in an
        array of shape labels_shape with the value of the index + 1
        corresponding to it, and 0 for background. For overlapping shapes
        z-ordering will be respected.

        Parameters
        ----------
        labels_shape : np.ndarray | tuple | None
            2-tuple defining shape of labels image to be generated. If non
            specified, takes the max of all the vertices
        zoom_factor : float
            Premultiplier applied to coordinates before generating mask. Used
            for generating as downsampled mask.
        offset : 2-tuple
            Offset subtracted from coordinates before multiplying by the
            zoom_factor. Used for putting negative coordinates into the mask.

        Returns
        -------
        labels : np.ndarray
            MxP integer array where each value is either 0 for background or an
            integer up to N for points inside the corresponding shape.
        """
        if labels_shape is None:
            labels_shape = self.displayed_vertices.max(axis=0).astype(np.int)

        labels = np.zeros(labels_shape, dtype=int)

        for ind in self._z_order[::-1]:
            mask = self.shapes[ind].to_mask(
                labels_shape, zoom_factor=zoom_factor, offset=offset
            )
            labels[mask] = ind + 1

        return labels

    def to_colors(
        self, colors_shape=None, zoom_factor=1, offset=[0, 0], max_shapes=None
    ):
        """Rasterize shapes to an RGBA image array.

        Each shape is embedded in an array of shape `colors_shape` with the
        RGBA value of the shape, and 0 for background. For overlapping shapes
        z-ordering will be respected.

        Parameters
        ----------
        colors_shape : np.ndarray | tuple | None
            2-tuple defining shape of colors image to be generated. If non
            specified, takes the max of all the vertiecs
        zoom_factor : float
            Premultiplier applied to coordinates before generating mask. Used
            for generating as downsampled mask.
        offset : 2-tuple
            Offset subtracted from coordinates before multiplying by the
            zoom_factor. Used for putting negative coordinates into the mask.
        max_shapes : None | int
            If provided, this is the maximum number of shapes that will be rasterized.
            If the number of shapes in view exceeds max_shapes, max_shapes shapes
            will be randomly selected from the in view shapes. If set to None, no
            maximum is applied. The default value is None.

        Returns
        -------
        colors : (N, M, 4) array
            rgba array where each value is either 0 for background or the rgba
            value of the shape for points inside the corresponding shape.
        """
        if colors_shape is None:
            colors_shape = self.displayed_vertices.max(axis=0).astype(np.int)

        colors = np.zeros(tuple(colors_shape) + (4,), dtype=float)
        colors[..., 3] = 1

        z_order = self._z_order[::-1]
        shapes_in_view = np.argwhere(self._displayed)
        z_order_in_view_mask = np.isin(z_order, shapes_in_view)
        z_order_in_view = z_order[z_order_in_view_mask]

        # If there are too many shapes to render responsively, just render
        # the top max_shapes shapes
        if max_shapes is not None and len(z_order_in_view) > max_shapes:
            z_order_in_view = z_order_in_view[0:max_shapes]

        for ind in z_order_in_view:
            mask = self.shapes[ind].to_mask(
                colors_shape, zoom_factor=zoom_factor, offset=offset
            )
            if type(self.shapes[ind]) in [Path, Line]:
                col = self._edge_color[ind]
            else:
                col = self._face_color[ind]
            colors[mask, :] = col

        return colors



from typing import TYPE_CHECKING

from ...utils.translations import trans
from ..dialogs.qt_about import QtAbout
from ._util import NapariMenu, populate_menu

if TYPE_CHECKING:
    from ..qt_main_window import Window

try:
    from napari_error_reporter import ask_opt_in
except ModuleNotFoundError:
    ask_opt_in = None


class HelpMenu(NapariMenu):
    def __init__(self, window: 'Window'):
        super().__init__(trans._('&Help'), window._qt_window)
        ACTIONS = [
            {
                'text': trans._('napari Info'),
                'slot': lambda e: QtAbout.showAbout(window._qt_window),
                'shortcut': 'Ctrl+/',
                'statusTip': trans._('About napari'),
            }
        ]
        if ask_opt_in is not None:
            ACTIONS.append(
                {
                    'text': trans._('Bug reporting opt in/out...'),
                    'slot': lambda: ask_opt_in(force=True),
                }
            )

        populate_menu(self, ACTIONS)



from typing import TYPE_CHECKING

import numpy as np

from ..utils._dtype import normalize_dtype
from ..utils.colormaps import ensure_colormap
from ..utils.events import Event
from ..utils.status_messages import format_float
from ..utils.validators import validate_n_seq

validate_2_tuple = validate_n_seq(2)

if TYPE_CHECKING:
    from .image.image import Image


class IntensityVisualizationMixin:
    """A mixin that adds gamma, colormap, and contrast limits logic to Layers.

    When used, this should come before the Layer in the inheritance, e.g.:

        class Image(IntensityVisualizationMixin, Layer):
            def __init__(self):
                ...
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.events.add(
            contrast_limits=Event,
            contrast_limits_range=Event,
            gamma=Event,
            colormap=Event,
        )
        self._gamma = 1
        self._colormap_name = ''
        self._contrast_limits_msg = ''
        self._contrast_limits = [None, None]
        self._contrast_limits_range = [None, None]
        self._auto_contrast_source = 'slice'
        self._keep_auto_contrast = False

    def reset_contrast_limits(self: 'Image', mode=None):
        """Scale contrast limits to data range"""
        mode = mode or self._auto_contrast_source
        self.contrast_limits = self._calc_data_range(mode)

    def reset_contrast_limits_range(self, mode=None):
        """Scale contrast limits range to data type if dtype is an integer,
        or use the current maximum data range otherwise.
        """
        dtype = normalize_dtype(self.dtype)
        if np.issubdtype(dtype, np.integer):
            info = np.iinfo(dtype)
            self.contrast_limits_range = (info.min, info.max)
        else:
            mode = mode or self._auto_contrast_source
            self.contrast_limits_range = self._calc_data_range(mode)

    @property
    def colormap(self):
        """napari.utils.Colormap: colormap for luminance images."""
        return self._colormap

    def _set_colormap(self, colormap):
        self._colormap = ensure_colormap(colormap)
        self._update_thumbnail()
        self.events.colormap()

    @colormap.setter
    def colormap(self, colormap):
        self._set_colormap(colormap)

    @property
    def colormaps(self):
        """tuple of str: names of available colormaps."""
        return tuple(self._colormaps.keys())

    @property
    def contrast_limits(self):
        """list of float: Limits to use for the colormap."""
        return list(self._contrast_limits)

    @contrast_limits.setter
    def contrast_limits(self, contrast_limits):
        validate_2_tuple(contrast_limits)
        self._contrast_limits_msg = (
            format_float(contrast_limits[0])
            + ', '
            + format_float(contrast_limits[1])
        )
        self._contrast_limits = contrast_limits
        # make sure range slider is big enough to fit range
        newrange = list(self.contrast_limits_range)
        newrange[0] = min(newrange[0], contrast_limits[0])
        newrange[1] = max(newrange[1], contrast_limits[1])
        self.contrast_limits_range = newrange
        self._update_thumbnail()
        self.events.contrast_limits()

    @property
    def contrast_limits_range(self):
        """The current valid range of the contrast limits."""
        return list(self._contrast_limits_range)

    @contrast_limits_range.setter
    def contrast_limits_range(self, value):
        """Set the valid range of the contrast limits"""
        validate_2_tuple(value)
        if list(value) == self.contrast_limits_range:
            return

        # if either value is "None", it just preserves the current range
        current_range = self.contrast_limits_range
        value = list(value)  # make sure it is mutable
        for i in range(2):
            value[i] = current_range[i] if value[i] is None else value[i]
        self._contrast_limits_range = value
        self.events.contrast_limits_range()

        # make sure that the current values fit within the new range
        # this also serves the purpose of emitting events.contrast_limits()
        # and updating the views/controllers
        if hasattr(self, '_contrast_limits') and any(self._contrast_limits):
            cur_min, cur_max = self.contrast_limits
            new_min = min(max(value[0], cur_min), value[1])
            new_max = max(min(value[1], cur_max), value[0])
            self.contrast_limits = (new_min, new_max)

    @property
    def gamma(self):
        return self._gamma

    @gamma.setter
    def gamma(self, value):
        self._gamma = value
        self._update_thumbnail()
        self.events.gamma()



from pathlib import Path
from typing import TYPE_CHECKING

import npe2
import numpy as np
import pytest
from conftest import LAYERS

from napari_builtins.io import napari_get_reader

if TYPE_CHECKING:
    from napari import layers

_EXTENSION_MAP = {
    'image': '.tif',
    'labels': '.tif',
    'points': '.csv',
    'shapes': '.csv',
}


@pytest.mark.parametrize('use_ext', [True, False])
def test_layer_save(tmp_path: Path, some_layer: 'layers.Layer', use_ext: bool):
    """Test saving layer data."""
    ext = _EXTENSION_MAP[some_layer._type_string]
    path_with_ext = tmp_path / f'layer_file{ext}'
    path_no_ext = tmp_path / 'layer_file'
    assert not path_with_ext.is_file()
    assert some_layer.save(str(path_with_ext if use_ext else path_no_ext))
    assert path_with_ext.is_file()

    # Read data back in
    reader = napari_get_reader(str(path_with_ext))
    assert callable(reader)
    [(read_data, *rest)] = reader(str(path_with_ext))

    if isinstance(some_layer.data, list):
        for d in zip(read_data, some_layer.data):
            np.testing.assert_allclose(*d)
    else:
        np.testing.assert_allclose(read_data, some_layer.data)

    if rest:
        meta, type_string = rest
        assert type_string == some_layer._type_string
        for key, value in meta.items():  # type: ignore
            np.testing.assert_equal(value, getattr(some_layer, key))


# the layer_writer_and_data fixture is defined in napari/conftest.py
def test_no_write_layer_bad_extension(some_layer: 'layers.Layer'):
    """Test not writing layer data with a bad extension."""
    with pytest.warns(UserWarning, match='No data written!'):
        assert not some_layer.save('layer.bad_extension')


# test_plugin_manager fixture is provided by napari_plugin_engine._testsupport
def test_get_writer_succeeds(tmp_path: Path):
    """Test writing layers data."""

    path = tmp_path / 'layers_folder'
    written = npe2.write(path=str(path), layer_data=LAYERS)  # type: ignore

    # check expected files were written
    expected = {
        str(path / f'{layer.name}{_EXTENSION_MAP[layer._type_string]}')
        for layer in LAYERS
    }
    assert path.is_dir()
    assert set(written) == expected
    for expect in expected:
        assert Path(expect).is_file()



import contextlib
from typing import TYPE_CHECKING, Callable, List, Union

from qtpy.QtWidgets import QAction, QMenu

if TYPE_CHECKING:
    from typing_extensions import TypedDict

    from ...utils.events import EventEmitter

    try:
        from qtpy.QtCore import SignalInstance
    except ImportError:
        from qtpy.QtCore import pyqtBoundSignal as SignalInstance

    class ActionDict(TypedDict):
        text: str
        # these are optional
        slot: Callable
        shortcut: str
        statusTip: str
        menuRole: QAction.MenuRole
        checkable: bool
        checked: bool
        check_on: Union[EventEmitter, SignalInstance]

    class MenuDict(TypedDict):
        menu: str
        # these are optional
        items: List[ActionDict]

    # note: TypedDict still doesn't have the concept of "optional keys"
    # so we add in generic `dict` for type checking.
    # see PEP655: https://peps.python.org/pep-0655/
    MenuItem = Union[MenuDict, ActionDict, dict]


def populate_menu(menu: QMenu, actions: List['MenuItem']):
    """Populate a QMenu from a declarative list of QAction dicts.

    Parameters
    ----------
    menu : QMenu
        the menu to populate
    actions : list of dict
        A list of dicts with one or more of the following keys

        **Required: One of "text" or "menu" MUST be present in the dict**
        text: str
            the name of the QAction to add
        menu: str
            if present, creates a submenu instead.  "menu" keys may also
            provide an "items" key to populate the menu.

        **Optional:**
        slot: callable
            a callback to call when the action is triggered
        shortcut: str
            a keyboard shortcut to trigger the actoin
        statusTip: str
            used for setStatusTip
        menuRole: QAction.MenuRole
            used for setMenuRole
        checkable: bool
            used for setCheckable
        checked: bool
            used for setChecked (only if `checkable` is provided and True)
        check_on: EventEmitter
            If provided, and `checkable` is True, this EventEmitter will be
            connected to action.setChecked:

            `dct['check_on'].connect(lambda e: action.setChecked(e.value))`
    """
    for ax in actions:
        if not ax:
            menu.addSeparator()
            continue
        if not ax.get("when", True):
            continue
        if 'menu' in ax:
            sub = ax['menu']
            if isinstance(sub, QMenu):
                menu.addMenu(sub)
                sub.setParent(menu)
            else:
                sub = menu.addMenu(sub)
            populate_menu(sub, ax.get("items", []))
            continue
        action: QAction = menu.addAction(ax['text'])
        if 'slot' in ax:
            if ax.get("checkable"):
                action.toggled.connect(ax['slot'])
            else:
                action.triggered.connect(ax['slot'])
        action.setShortcut(ax.get('shortcut', ''))
        action.setStatusTip(ax.get('statusTip', ''))
        if 'menuRole' in ax:
            action.setMenuRole(ax['menuRole'])
        if ax.get("checkable"):
            action.setCheckable(True)
            action.setChecked(ax.get("checked", False))
            if 'check_on' in ax:
                emitter = ax['check_on']

                @emitter.connect
                def _setchecked(e, action=action):
                    action.setChecked(e.value if hasattr(e, 'value') else e)

        action.setData(ax)


class NapariMenu(QMenu):
    """
    Base napari menu class that provides action handling and clean up on
    close.
    """

    _INSTANCES: List['NapariMenu'] = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._INSTANCES.append(self)

    def _destroy(self):
        """Clean up action data to avoid widget leaks."""
        for ax in self.actions():
            ax.setData(None)

            with contextlib.suppress(AttributeError):
                ax._destroy()
        if self in self._INSTANCES:
            self._INSTANCES.remove(self)

    def update(self, event=None):
        """Update action enabled/disabled state based on action data."""
        for ax in self.actions():
            if data := ax.data():
                enabled_func = data.get('enabled', lambda event: True)
                ax.setEnabled(bool(enabled_func(event)))



from enum import auto

from ...utils.misc import StringEnum


class Anchor(StringEnum):
    """
    Anchor: The anchor position for text

    CENTER The text origin is centered on the layer item bounding box.

    UPPER_LEFT The text origin is on the upper left corner of the bounding box
    UPPER_RIGHT The text origin is on the upper right corner of the bounding box
    LOWER_LEFT The text origin is on the lower left corner of the bounding box
    LOWER_RIGHT The text origin is on the lower right corner of the bounding box
    """

    CENTER = auto()
    UPPER_LEFT = auto()
    UPPER_RIGHT = auto()
    LOWER_LEFT = auto()
    LOWER_RIGHT = auto()



"""OctreeImage class.

An eventual replacement for Image that combines single-scale and
chunked (tiled) multi-scale into one implementation.
"""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, List, Set

import numpy as np

from ....utils.events import Event
from ....utils.translations import trans
from ..image import _ImageBase
from ._octree_slice import OctreeSlice, OctreeView
from .octree_chunk import OctreeChunk
from .octree_intersection import OctreeIntersection
from .octree_level import OctreeLevelInfo
from .octree_util import OctreeDisplayOptions, OctreeMetadata

if TYPE_CHECKING:
    from ....components.experimental.chunk import ChunkRequest

LOGGER = logging.getLogger("napari.octree.image")


class _OctreeImageBase(_ImageBase):
    """Image layer rendered using an octree.

    Experimental variant of Image that renders using an octree. For 2D
    images the octree is really just a quadtree. For 3D volumes it will be
    a real octree. This class is intended to eventually replace the
    existing Image class.

    Notes
    -----
    The original Image class handled single-scale and multi-scale images,
    but they were handled quite differently. And its multi-scale did not
    use chunks or tiles. It worked well on local data, but was basically
    unusable for remote or high latency data.

    OctreeImage always uses chunk/tiles. Today those tiles are always
    "small". However, as a special case, if an image is smaller than the
    max texture size, we could some day allow OctreeImage to set its tile
    size equal to that image size.

    At that point "small" images would be draw with a single texture,
    the same way the old Image class drew then. So it would be very
    efficient.

    But larger images would have multiple chunks/tiles and multiple levels.
    Unlike the original Image class multi-scale, the chunks/tiles mean we
    only have to incrementally load more data as the user pans and zooms.

    The goal is OctreeImage gets renamed to just Image and it efficiently
    handles images of any size. It make take a while to get there.

    Attributes
    ----------
    _view : OctreeView
        Describes a view frustum which implies what portion of the OctreeImage
        needs to be draw.
    _slice : OctreeSlice
        When _set_view_slice() is called we create a OctreeSlice()
        that's looking at some specific slice of the data.
    _display : OctreeDisplayOptions
        Settings for how we draw the octree, such as tile size.
    """

    def __init__(self, *args, **kwargs):

        self._view: OctreeView = None
        self._slice: OctreeSlice = None
        self._intersection: OctreeIntersection = None
        self._display = OctreeDisplayOptions()

        # super().__init__ will call our _set_view_slice() which is kind
        # of annoying since we are aren't fully constructed yet.
        super().__init__(*args, **kwargs)

        # Call after super().__init__
        self.events.add(octree_level=Event, tile_size=Event)

        # TODO_OCTREE: this is hack that we assign OctreeDisplayOptions
        # this event after super().__init__(). Needs to be cleaned up.
        self._display.loaded_event = self.events.loaded

    def _get_value(self, position):
        """Override Image._get_value(position)."""
        return (0, (0, 0))  # TODO_OCTREE: need to implement this.

    @property
    def loaded(self) -> bool:
        """Has the data for this layer been loaded yet.

        As far as the visual system is concerned we are always "loaded" in
        that we can always be drawn. Because our VispyTiledImageLayer can
        always be drawn. Even if no chunk/tiles are loaded yet.
        """
        return True

    @property
    def _empty(self) -> bool:
        """Is this layer completely empty so it can't be drawn.

        As with self.loaded, we are never really empty. Our VispyTiledImageLayer
        can always be drawn. Even if there is nothing to draw.
        """
        return False

    def _update_thumbnail(self):
        # TODO_OCTREE: replace Image._update_thumbnail with nothing for
        # the moment until we decide how to do thumbnail.
        pass

    @property
    def _data_view(self):
        """Viewable image for the current slice. (compatibility)"""
        # Override Image._data_view
        return np.zeros((64, 64, 3))  # fake: does octree need this?

    @property
    def display(self) -> OctreeDisplayOptions:
        """The display options for this octree image layer."""
        return self._display

    @property
    def tile_size(self) -> int:
        """Return the edge length of single tile, for example 256.

        Returns
        -------
        int
            The edge length of a single tile.
        """
        return self._display.tile_size

    @tile_size.setter
    def tile_size(self, tile_size: int) -> None:
        """Set new tile_size.

        Parameters
        ----------
        tile_size : int
            The new tile size.
        """
        self._display.tile_size = tile_size
        self.events.tile_size()

        self._slice = None  # For now must explicitly delete it
        self.refresh()  # Creates a new slice.

    @property
    def tile_shape(self) -> tuple:
        """Return the shape of a single tile, for example 256x256x3.

        Returns
        -------
        tuple
            The shape of a single tile.
        """
        if self.multiscale:
            init_shape = self.data[0].shape
        else:
            init_shape = self.data.shape

        tile_shape = (self.tile_size, self.tile_size)

        if self.rgb:
            # Add the color dimension (usually 3 or 4)
            tile_shape += (init_shape[-1],)

        return tile_shape

    @property
    def meta(self) -> OctreeMetadata:
        """Information about the current octree.

        Returns
        -------
        OctreeMetadata
            Octree dimensions and other info.
        """
        if self._slice is None:
            return None
        return self._slice.meta

    @property
    def octree_level_info(self) -> OctreeLevelInfo:
        """Information about the current level of the current octree.

        Returns
        -------
        OctreeLevelInfo
            Information about the current octree level.
        """
        if self._slice is None:
            return None
        return self._slice.octree_level_info

    @property
    def data_level(self) -> int:
        """Current level of multiscale.

        The base full resolution image is level 0. The highest and coarsest
        level usually contains only a single tile.
        """
        return self._data_level

    @data_level.setter
    def data_level(self, level: int) -> None:
        """Set the octree level we should be displaying.

        Parameters
        ----------
        level : int
            Display this octree level.
        """
        if self._data_level == level:
            return  # It didn't change.

        # Quickly check for less than 0. We can't check for a level
        # that's too high because the Octree might have extended levels?
        if level < 0:
            raise ValueError(
                trans._(
                    "Octree level {level} is negative.",
                    deferred=True,
                    level=level,
                )
            )

        self._data_level = level
        self.events.octree_level()

        if self._slice is not None:
            # This will raise if the level is too high.
            self._slice.octree_level = level

        self.events.loaded()  # redraw

    @property
    def num_octree_levels(self) -> int:
        """Return the total number of octree levels.

        Returns
        -------
        int
            The number of octree levels.
        """
        return len(self.data)  # Multiscale

    def _new_empty_slice(self) -> None:
        """Initialize the current slice to an empty image.

        Overides Image._new_empty_slice() and does nothing because we don't
        need an empty slice. We create self._slice when
        self._set_view_slice() is called.

        The empty slice was needed to satisfy the old VispyImageLayer that
        used a single ImageVisual. But OctreeImage is drawn with
        VispyTiledImageVisual. It does not need an empty image. It gets
        chunks from our self.drawable_chunks property, and it will just draw
        nothing if that returns an empty list.

        When OctreeImage become the only image class, this can go away.
        """

    def get_drawable_chunks(
        self, drawn_set: Set[OctreeChunk]
    ) -> List[OctreeChunk]:
        """Get the chunks in the current slice which are drawable.

        The visual calls this and then draws what we send it. The call to
        get_intersection() will chose the appropriate level of the octree
        to intersect, and then return all the chunks within the
        intersection with that level.

        These are the "ideal" chunks because they are at the level whose
        resolution best matches the current screen resolution.

        Drawing chunks at a lower level than this will work fine, but it's
        a waste in that those chunks will just be downsampled by the card.
        You won't see any "extra" resolution at all. The card can do this
        super fast, so the issue not such much speed as it is RAM and VRAM.

        In the opposite direction, drawing chunks from a higher, the number
        of chunks and storage goes down quickly. The only issue there is
        visual quality, the imagery might look blurry.

        Parameters
        ----------
        drawn_set : Set[OctreeChunk]
            The chunks that are currently being drawn by the visual.

        Returns
        -------
        List[OctreeChunk]
            The drawable chunks.
        """
        if self._slice is None or self._view is None:
            LOGGER.debug("get_drawable_chunks: No slice or view")
            return []  # There is nothing to draw.

        # TODO_OCTREE: Make this a config option, maybe different
        # expansion_factor each level above the ideal level?
        expansion_factor = 1.1
        view = self._view.expand(expansion_factor)

        # Get the current intersection and save it off.
        self._intersection = self._slice.get_intersection(view)

        if self._intersection is None:
            LOGGER.debug("get_drawable_chunks: Intersection is empty")
            return []  # No chunks to draw.

        # Get the ideal chunks. These are the chunks at the preferred
        # resolution. The ones we ideally want to draw once they are in RAM
        # and in VRAM. When all loading is done, we will draw all the ideal
        # chunks.
        ideal_chunks = self._intersection.get_chunks(create=True)
        ideal_level = self._intersection.level.info.level_index

        # log_chunks("ideal_chunks", ideal_chunks)

        # If we are seting the data level level automatically, then update
        # our level to match what was chosen for the intersection.
        if self._view.auto_level:
            self._data_level = ideal_level

        # The loader will initiate loads on any ideal chunks which are not
        # yet in memory. And it will return the chunks we should draw. The
        # chunks we should draw might be ideal chunks, if they are in
        # memory, but they also might be chunks from higher or lower levels
        # in the octree. In general we try to draw "cover the view" with
        # the "best available" data.
        return self._slice.loader.get_drawable_chunks(
            drawn_set, ideal_chunks, ideal_level
        )

    def _update_draw(
        self, scale_factor, corner_pixels_displayed, shape_threshold
    ) -> None:
        """Override Layer._update_draw completely.

        The base Layer._update_draw does stuff for the legacy multi-scale
        that we don't want. And it calls refresh() which we don't need.

        We create our OctreeView() here which has the corners in it.

        Parameters
        ----------
        scale_factor : float
            Scale factor going from canvas to world coordinates.
        corner_pixels_displayed : array
            Coordinates of the top-left and bottom-right canvas pixels in
            world coordinates.
        shape_threshold : tuple
            Requested shape of field of view in data coordinates.

        """
        # Compute our 2D corners from the incoming n-d corner_pixels
        data_corners = (
            self._transforms[1:]
            .simplified.set_slice(self._displayed_axes)
            .inverse(corner_pixels_displayed)
        )

        # Update our self._view to to capture the state of things right
        # before we are drawn. Our self._view will used by our
        # drawable_chunks() method.
        self._view = OctreeView(data_corners, shape_threshold, self.display)

    def get_intersection(self) -> OctreeIntersection:
        """The the interesection between the current view and the octree.

        Returns
        -------
        OctreeIntersection
            The intersection between the current view and the octree.
        """
        if self._slice is None:
            return None

        return self._slice.get_intersection(self._view)

    def _outside_data_range(self, indices) -> bool:
        """Return True if requested slice is outside of data range.

        Returns
        -------
        bool
            True if requested slice is outside data range.
        """

        extent = self._extent_data
        not_disp = self._dims_not_displayed

        return np.any(
            np.less(
                [indices[ax] for ax in not_disp],
                [extent[0, ax] for ax in not_disp],
            )
        ) or np.any(
            np.greater(
                [indices[ax] for ax in not_disp],
                [extent[1, ax] for ax in not_disp],
            )
        )

    def _set_view_slice(self) -> None:
        """Set the view given the indices to slice with.

        This replaces Image._set_view_slice() entirely. The hope is eventually
        this class OctreeImage becomes Image. And the non-tiled multiscale
        logic in Image._set_view_slice goes away entirely.
        """
        # Consider non-multiscale data as just having a single level
        from ....components.experimental.chunk import LayerRef

        multilevel_data = self.data if self.multiscale else [self.data]

        if self._slice is not None:
            # For now bail out so we don't nuke an existing slice which
            # contains an existing octree. Soon we'll need to figure out
            # if we are really changing slices (and need a new octree).
            return

        indices = np.array(self._slice_indices)
        if self._outside_data_range(indices):
            return

        # Indices to get at the data we are currently viewing.
        indices = self._get_slice_indices()

        # TODO_OCTREE: easier way to do this?
        base_shape = multilevel_data[0].shape
        base_shape_2d = [base_shape[i] for i in self._dims_displayed]

        layer_ref = LayerRef.from_layer(self)

        meta = OctreeMetadata(
            layer_ref,
            base_shape_2d,
            len(multilevel_data),
            self._display.tile_size,
        )

        # OctreeSlice wants all the levels, but only the dimensions
        # of each level that we are currently viewing.
        slice_data = [level_data[indices] for level_data in multilevel_data]
        layer_ref = LayerRef.from_layer(self)

        # Create the slice, it will create the actual Octree.
        self._slice = OctreeSlice(
            slice_data,
            layer_ref,
            meta,
        )

    def _get_slice_indices(self) -> tuple:
        """Get the slice indices including possible depth for RGB."""
        indices = tuple(self._slice_indices)

        if self.rgb:
            indices += (slice(None),)

        return indices

    def on_chunk_loaded(self, request: ChunkRequest) -> None:
        """An asynchronous ChunkRequest was loaded.

        Override Image.on_chunk_loaded() fully.

        Parameters
        ----------
        request : ChunkRequest
            This request was loaded.
        """
        LOGGER.info(
            "on_chunk_loaded: load=%.3fms elapsed=%.3fms location = %s",
            request.load_ms,
            request.elapsed_ms,
            request.location,
        )

        # Pass it to the slice, it will insert the newly loaded data into
        # the OctreeChunk at the right location.
        if self._slice.on_chunk_loaded(request):
            # Redraw with the new chunk.
            # TODO_OCTREE: Call this at most once per frame? It's a bad
            # idea to call it for every chunk?
            LOGGER.debug("on_chunk_loaded calling loaded()")
            self.events.loaded()

    @property
    def remote_messages(self) -> dict:
        """Messages we should send to remote clients."""
        if self._intersection is None:
            return {}

        return {
            "tile_state": self._intersection.tile_state,
            "tile_config": self._intersection.tile_config,
        }
